import ast
import sys
from ast import *  # noqa: F401,F403

from rope.base import fscommands

try:
    # Suppress the mypy complaint: Module "ast" has no attribute "_const_node_type_names"
    from ast import _const_node_type_names  # type:ignore
except ImportError:
    # backported from stdlib `ast`
    assert sys.version_info < (3, 8)
    _const_node_type_names = {
        bool: "NameConstant",  # should be before int
        type(None): "NameConstant",
        int: "Num",
        float: "Num",
        complex: "Num",
        str: "Str",
        bytes: "Bytes",
        type(...): "Ellipsis",
    }


def parse(source, filename="<string>", *args, **kwargs):  # type: ignore
    if isinstance(source, str):
        source = fscommands.unicode_to_file_data(source)
    if b"\r" in source:
        source = source.replace(b"\r\n", b"\n").replace(b"\r", b"\n")
    if not source.endswith(b"\n"):
        source += b"\n"
    try:
        return ast.parse(source, filename=filename, *args, **kwargs)
    except (TypeError, ValueError) as e:
        error = SyntaxError()
        error.lineno = 1
        error.filename = filename
        error.msg = str(e)
        raise error


def call_for_nodes(node, callback):
    """
    Pre-order depth-first traversal of AST nodes, calling `callback(node)` for
    each node visited.

    When each node is visited, `callback(node)` will be called with the visited
    `node`, then its children node will be visited.

    If `callback(node)` returns `True` for a node, then the descendants of that
    node will not be visited.

    See _ResultChecker._find_node for an example.
    """
    result = callback(node)
    if not result:
        for child in ast.iter_child_nodes(node):
            call_for_nodes(child, callback)


class RopeNodeVisitor(ast.NodeVisitor):
    def visit(self, node):
        """Modified from ast.NodeVisitor to match rope's existing Visitor implementation"""
        method = "_" + node.__class__.__name__
        visitor = getattr(self, method, self.generic_visit)
        return visitor(node)


def get_const_subtype_name(node):
    """Get pre-3.8 ast node name"""
    # fmt: off
    assert sys.version_info >= (3, 8), "This should only be called in Python 3.8 and above"
    # fmt: on
    assert isinstance(node, ast.Constant)
    return _const_node_type_names[type(node.value)]


def get_node_type_name(node):
    return (
        get_const_subtype_name(node)
        if isinstance(node, ast.Constant)
        else node.__class__.__name__
    )



import unittest
from textwrap import dedent

import rope.base.codeanalyze
import rope.base.exceptions
from rope.refactor import extract
from ropetest import testutils


class ExtractMethodTest(unittest.TestCase):
    def setUp(self):
        super().setUp()
        self.project = testutils.sample_project()
        self.pycore = self.project.pycore

    def tearDown(self):
        testutils.remove_project(self.project)
        super().tearDown()

    def do_extract_method(self, source_code, start, end, extracted, **kwds):
        testmod = testutils.create_module(self.project, "testmod")
        testmod.write(source_code)
        extractor = extract.ExtractMethod(self.project, testmod, start, end)
        self.project.do(extractor.get_changes(extracted, **kwds))
        return testmod.read()

    def do_extract_variable(self, source_code, start, end, extracted, **kwds):
        testmod = testutils.create_module(self.project, "testmod")
        testmod.write(source_code)
        extractor = extract.ExtractVariable(self.project, testmod, start, end)
        self.project.do(extractor.get_changes(extracted, **kwds))
        return testmod.read()

    def _convert_line_range_to_offset(self, code, start, end):
        lines = rope.base.codeanalyze.SourceLinesAdapter(code)
        return lines.get_line_start(start), lines.get_line_end(end)

    def test_simple_extract_function(self):
        code = dedent("""\
            def a_func():
                print('one')
                print('two')
        """)
        start, end = self._convert_line_range_to_offset(code, 2, 2)
        refactored = self.do_extract_method(code, start, end, "extracted")
        expected = dedent("""\
            def a_func():
                extracted()
                print('two')

            def extracted():
                print('one')
        """)
        self.assertEqual(expected, refactored)

    def test_simple_extract_function_one_line(self):
        code = dedent("""\
            def a_func():
                resp = 'one'
                print(resp)
        """)
        selected = "'one'"
        start, end = code.index(selected), code.index(selected) + len(selected)
        refactored = self.do_extract_method(code, start, end, "extracted")
        expected = dedent("""\
            def a_func():
                resp = extracted()
                print(resp)

            def extracted():
                return 'one'
        """)
        self.assertEqual(expected, refactored)

    def test_extract_function_at_the_end_of_file(self):
        code = dedent("""\
            def a_func():
                print('one')""")
        start, end = self._convert_line_range_to_offset(code, 2, 2)
        refactored = self.do_extract_method(code, start, end, "extracted")
        expected = dedent("""\
            def a_func():
                extracted()
            def extracted():
                print('one')
        """)
        self.assertEqual(expected, refactored)

    def test_extract_function_after_scope(self):
        code = dedent("""\
            def a_func():
                print('one')
                print('two')

            print('hey')
        """)
        start, end = self._convert_line_range_to_offset(code, 2, 2)
        refactored = self.do_extract_method(code, start, end, "extracted")
        expected = dedent("""\
            def a_func():
                extracted()
                print('two')

            def extracted():
                print('one')

            print('hey')
        """)
        self.assertEqual(expected, refactored)

    def test_extract_function_containing_dict_generalized_unpacking(self):
        code = dedent("""\
            def a_func(dict1):
                dict2 = {}
                a_var = {a: b, **dict1, **dict2}
        """)
        start = code.index("{a")
        end = code.index("2}") + len("2}")
        refactored = self.do_extract_method(code, start, end, "extracted")
        expected = dedent("""\
            def a_func(dict1):
                dict2 = {}
                a_var = extracted(dict1, dict2)

            def extracted(dict1, dict2):
                return {a: b, **dict1, **dict2}
        """)
        self.assertEqual(expected, refactored)

    def test_simple_extract_function_with_parameter(self):
        code = dedent("""\
            def a_func():
                a_var = 10
                print(a_var)
        """)
        start, end = self._convert_line_range_to_offset(code, 3, 3)
        refactored = self.do_extract_method(code, start, end, "new_func")
        expected = dedent("""\
            def a_func():
                a_var = 10
                new_func(a_var)

            def new_func(a_var):
                print(a_var)
        """)
        self.assertEqual(expected, refactored)

    def test_not_unread_variables_as_parameter(self):
        code = dedent("""\
            def a_func():
                a_var = 10
                print('hey')
        """)
        start, end = self._convert_line_range_to_offset(code, 3, 3)
        refactored = self.do_extract_method(code, start, end, "new_func")
        expected = dedent("""\
            def a_func():
                a_var = 10
                new_func()

            def new_func():
                print('hey')
        """)
        self.assertEqual(expected, refactored)

    def test_simple_extract_function_with_two_parameter(self):
        code = dedent("""\
            def a_func():
                a_var = 10
                another_var = 20
                third_var = a_var + another_var
        """)
        start, end = self._convert_line_range_to_offset(code, 4, 4)
        refactored = self.do_extract_method(code, start, end, "new_func")
        expected = dedent("""\
            def a_func():
                a_var = 10
                another_var = 20
                new_func(a_var, another_var)

            def new_func(a_var, another_var):
                third_var = a_var + another_var
        """)
        self.assertEqual(expected, refactored)

    def test_simple_extract_function_with_return_value(self):
        code = dedent("""\
            def a_func():
                a_var = 10
                print(a_var)
        """)
        start, end = self._convert_line_range_to_offset(code, 2, 2)
        refactored = self.do_extract_method(code, start, end, "new_func")
        expected = dedent("""\
            def a_func():
                a_var = new_func()
                print(a_var)

            def new_func():
                a_var = 10
                return a_var
        """)
        self.assertEqual(expected, refactored)

    def test_extract_function_with_kwonlyargs(self):
        code = dedent("""\
            def a_func(b, *, a_var):
                another_var = 20
                third_var = a_var + another_var
        """)
        start, end = self._convert_line_range_to_offset(code, 3, 3)
        refactored = self.do_extract_method(code, start, end, "new_func")
        expected = dedent("""\
            def a_func(b, *, a_var):
                another_var = 20
                new_func(a_var, another_var)

            def new_func(a_var, another_var):
                third_var = a_var + another_var
        """)
        self.assertEqual(expected, refactored)

    @testutils.only_for_versions_higher("3.8")
    def test_extract_function_with_posonlyargs(self):
        code = dedent("""\
            def a_func(a_var, /, b):
                another_var = 20
                third_var = a_var + another_var
        """)
        start, end = self._convert_line_range_to_offset(code, 3, 3)
        refactored = self.do_extract_method(code, start, end, "new_func")
        expected = dedent("""\
            def a_func(a_var, /, b):
                another_var = 20
                new_func(a_var, another_var)

            def new_func(a_var, another_var):
                third_var = a_var + another_var
        """)
        self.assertEqual(expected, refactored)

    def test_extract_function_with_multiple_return_values(self):
        code = dedent("""\
            def a_func():
                a_var = 10
                another_var = 20
                third_var = a_var + another_var
        """)
        start, end = self._convert_line_range_to_offset(code, 2, 3)
        refactored = self.do_extract_method(code, start, end, "new_func")
        expected = dedent("""\
            def a_func():
                a_var, another_var = new_func()
                third_var = a_var + another_var

            def new_func():
                a_var = 10
                another_var = 20
                return a_var, another_var
        """)
        self.assertEqual(expected, refactored)

    def test_simple_extract_method(self):
        code = dedent("""\
            class AClass(object):

                def a_func(self):
                    print(1)
                    print(2)
        """)
        start, end = self._convert_line_range_to_offset(code, 4, 4)
        refactored = self.do_extract_method(code, start, end, "new_func")
        expected = dedent("""\
            class AClass(object):

                def a_func(self):
                    self.new_func()
                    print(2)

                def new_func(self):
                    print(1)
        """)
        self.assertEqual(expected, refactored)

    def test_extract_method_with_args_and_returns(self):
        code = dedent("""\
            class AClass(object):
                def a_func(self):
                    a_var = 10
                    another_var = a_var * 3
                    third_var = a_var + another_var
        """)
        start, end = self._convert_line_range_to_offset(code, 4, 4)
        refactored = self.do_extract_method(code, start, end, "new_func")
        expected = dedent("""\
            class AClass(object):
                def a_func(self):
                    a_var = 10
                    another_var = self.new_func(a_var)
                    third_var = a_var + another_var

                def new_func(self, a_var):
                    another_var = a_var * 3
                    return another_var
        """)
        self.assertEqual(expected, refactored)

    def test_extract_method_args_and_kwargs(self):
        code = dedent("""\
            def a_func(a, *args, **kwargs):
                print(a, args, kwargs)
        """)
        start, end = self._convert_line_range_to_offset(code, 2, 2)
        refactored = self.do_extract_method(code, start, end, "new_func")
        expected = dedent("""\
            def a_func(a, *args, **kwargs):
                new_func(a, args, kwargs)

            def new_func(a, args, kwargs):
                print(a, args, kwargs)
        """)
        self.assertEqual(expected, refactored)

    def test_extract_method_with_self_as_argument(self):
        code = dedent("""\
            class AClass(object):
                def a_func(self):
                    print(self)
        """)
        start, end = self._convert_line_range_to_offset(code, 3, 3)
        refactored = self.do_extract_method(code, start, end, "new_func")
        expected = dedent("""\
            class AClass(object):
                def a_func(self):
                    self.new_func()

                def new_func(self):
                    print(self)
        """)
        self.assertEqual(expected, refactored)

    def test_extract_method_with_no_self_as_argument(self):
        code = dedent("""\
            class AClass(object):
                def a_func():
                    print(1)
        """)
        start, end = self._convert_line_range_to_offset(code, 3, 3)
        with self.assertRaises(rope.base.exceptions.RefactoringError):
            self.do_extract_method(code, start, end, "new_func")

    def test_extract_method_with_multiple_methods(self):
        code = dedent("""\
            class AClass(object):
                def a_func(self):
                    print(self)

                def another_func(self):
                    pass
        """)
        start, end = self._convert_line_range_to_offset(code, 3, 3)
        refactored = self.do_extract_method(code, start, end, "new_func")
        expected = dedent("""\
            class AClass(object):
                def a_func(self):
                    self.new_func()

                def new_func(self):
                    print(self)

                def another_func(self):
                    pass
        """)
        self.assertEqual(expected, refactored)

    def test_extract_function_with_function_returns(self):
        code = dedent("""\
            def a_func():
                def inner_func():
                    pass
                inner_func()
        """)
        start, end = self._convert_line_range_to_offset(code, 2, 3)
        refactored = self.do_extract_method(code, start, end, "new_func")
        expected = dedent("""\
            def a_func():
                inner_func = new_func()
                inner_func()

            def new_func():
                def inner_func():
                    pass
                return inner_func
        """)
        self.assertEqual(expected, refactored)

    def test_simple_extract_global_function(self):
        code = dedent("""\
            print('one')
            print('two')
            print('three')
        """)
        start, end = self._convert_line_range_to_offset(code, 2, 2)
        refactored = self.do_extract_method(code, start, end, "new_func")
        expected = dedent("""\
            print('one')

            def new_func():
                print('two')

            new_func()
            print('three')
        """)
        self.assertEqual(expected, refactored)

    def test_extract_global_function_inside_ifs(self):
        code = dedent("""\
            if True:
                a = 10
        """)
        start, end = self._convert_line_range_to_offset(code, 2, 2)
        refactored = self.do_extract_method(code, start, end, "new_func")
        expected = dedent("""\

            def new_func():
                a = 10

            if True:
                new_func()
        """)
        self.assertEqual(expected, refactored)

    def test_extract_function_while_inner_function_reads(self):
        code = dedent("""\
            def a_func():
                a_var = 10
                def inner_func():
                    print(a_var)
                return inner_func
        """)
        start, end = self._convert_line_range_to_offset(code, 3, 4)
        refactored = self.do_extract_method(code, start, end, "new_func")
        expected = dedent("""\
            def a_func():
                a_var = 10
                inner_func = new_func(a_var)
                return inner_func

            def new_func(a_var):
                def inner_func():
                    print(a_var)
                return inner_func
        """)
        self.assertEqual(expected, refactored)

    def test_extract_method_bad_range(self):
        code = dedent("""\
            def a_func():
                pass
            a_var = 10
        """)
        start, end = self._convert_line_range_to_offset(code, 2, 3)
        with self.assertRaises(rope.base.exceptions.RefactoringError):
            self.do_extract_method(code, start, end, "new_func")

    def test_extract_method_bad_range2(self):
        code = dedent("""\
            class AClass(object):
                pass
        """)
        start, end = self._convert_line_range_to_offset(code, 1, 1)
        with self.assertRaises(rope.base.exceptions.RefactoringError):
            self.do_extract_method(code, start, end, "new_func")

    def test_extract_method_containing_return(self):
        code = dedent("""\
            def a_func(arg):
                if arg:
                    return arg * 2
                return 1""")
        start, end = self._convert_line_range_to_offset(code, 2, 4)
        with self.assertRaises(rope.base.exceptions.RefactoringError):
            self.do_extract_method(code, start, end, "new_func")

    def test_extract_method_containing_yield(self):
        code = dedent("""\
            def a_func(arg):
                yield arg * 2
        """)
        start, end = self._convert_line_range_to_offset(code, 2, 2)
        with self.assertRaises(rope.base.exceptions.RefactoringError):
            self.do_extract_method(code, start, end, "new_func")

    def test_extract_method_containing_incomplete_lines(self):
        code = dedent("""\
            a_var = 20
            another_var = 30
        """)
        start = code.index("20")
        end = code.index("30") + 2
        with self.assertRaises(rope.base.exceptions.RefactoringError):
            self.do_extract_method(code, start, end, "new_func")

    def test_extract_method_containing_incomplete_lines2(self):
        code = dedent("""\
            a_var = 20
            another_var = 30
        """)
        start = code.index("20")
        end = code.index("another") + 5
        with self.assertRaises(rope.base.exceptions.RefactoringError):
            self.do_extract_method(code, start, end, "new_func")

    def test_extract_function_and_argument_as_parameter(self):
        code = dedent("""\
            def a_func(arg):
                print(arg)
        """)
        start, end = self._convert_line_range_to_offset(code, 2, 2)
        refactored = self.do_extract_method(code, start, end, "new_func")
        expected = dedent("""\
            def a_func(arg):
                new_func(arg)

            def new_func(arg):
                print(arg)
        """)
        self.assertEqual(expected, refactored)

    def test_extract_function_and_end_as_the_start_of_a_line(self):
        code = dedent("""\
            print("hey")
            if True:
                pass
        """)
        start = 0
        end = code.index("\n") + 1
        refactored = self.do_extract_method(code, start, end, "new_func")
        expected = dedent("""\

            def new_func():
                print("hey")

            new_func()
            if True:
                pass
        """)
        self.assertEqual(expected, refactored)

    def test_extract_function_and_indented_blocks(self):
        code = dedent("""\
            def a_func(arg):
                if True:
                    if True:
                        print(arg)
        """)
        start, end = self._convert_line_range_to_offset(code, 3, 4)
        refactored = self.do_extract_method(code, start, end, "new_func")
        expected = dedent("""\
            def a_func(arg):
                if True:
                    new_func(arg)

            def new_func(arg):
                if True:
                    print(arg)
        """)
        self.assertEqual(expected, refactored)

    def test_extract_method_and_multi_line_headers(self):
        code = dedent("""\
            def a_func(
                       arg):
                print(arg)
        """)
        start, end = self._convert_line_range_to_offset(code, 3, 3)
        refactored = self.do_extract_method(code, start, end, "new_func")
        expected = dedent("""\
            def a_func(
                       arg):
                new_func(arg)

            def new_func(arg):
                print(arg)
        """)
        self.assertEqual(expected, refactored)

    def test_single_line_extract_function(self):
        code = dedent("""\
            a_var = 10 + 20
        """)
        start = code.index("10")
        end = code.index("20") + 2
        refactored = self.do_extract_method(code, start, end, "new_func")
        expected = dedent("""\

            def new_func():
                return 10 + 20

            a_var = new_func()
        """)
        self.assertEqual(expected, refactored)

    def test_single_line_extract_function2(self):
        code = dedent("""\
            def a_func():
                a = 10
                b = a * 20
        """)
        start = code.rindex("a")
        end = code.index("20") + 2
        refactored = self.do_extract_method(code, start, end, "new_func")
        expected = dedent("""\
            def a_func():
                a = 10
                b = new_func(a)

            def new_func(a):
                return a * 20
        """)
        self.assertEqual(expected, refactored)

    def test_single_line_extract_method_and_logical_lines(self):
        code = dedent("""\
            a_var = 10 +\\
                20
        """)
        start = code.index("10")
        end = code.index("20") + 2
        refactored = self.do_extract_method(code, start, end, "new_func")
        expected = dedent("""\

            def new_func():
                return 10 + 20

            a_var = new_func()
        """)
        self.assertEqual(expected, refactored)

    def test_single_line_extract_method_and_logical_lines2(self):
        code = dedent("""\
            a_var = (10,\\
                20)
        """)
        start = code.index("10") - 1
        end = code.index("20") + 3
        refactored = self.do_extract_method(code, start, end, "new_func")
        expected = dedent("""\

            def new_func():
                return (10, 20)

            a_var = new_func()
        """)
        self.assertEqual(expected, refactored)

    def test_single_line_extract_method_with_large_multiline_expression(self):
        code = dedent("""\
            a_var = func(
                {
                    "hello": 1,
                    "world": 2,
                },
                blah=foo,
            )
        """)
        start = code.index("{") - 1
        end = code.index("}") + 1
        refactored = self.do_extract_method(code, start, end, "new_func")
        expected = dedent("""\

            def new_func():
                return {
                    "hello": 1,
                    "world": 2,
                }

            a_var = func(
                new_func(),
                blah=foo,
            )
        """)
        self.assertEqual(expected, refactored)

    def test_single_line_extract_method(self):
        code = dedent("""\
            class AClass(object):

                def a_func(self):
                    a = 10
                    b = a * a
        """)
        start = code.rindex("=") + 2
        end = code.rindex("a") + 1
        refactored = self.do_extract_method(code, start, end, "new_func")
        expected = dedent("""\
            class AClass(object):

                def a_func(self):
                    a = 10
                    b = self.new_func(a)

                def new_func(self, a):
                    return a * a
        """)
        self.assertEqual(expected, refactored)

    def test_single_line_extract_function_if_condition(self):
        code = dedent("""\
            if True:
                pass
        """)
        start = code.index("True")
        end = code.index("True") + 4
        refactored = self.do_extract_method(code, start, end, "new_func")
        expected = dedent("""\

            def new_func():
                return True

            if new_func():
                pass
        """)
        self.assertEqual(expected, refactored)

    def test_unneeded_params(self):
        code = dedent("""\
            class A(object):
                def a_func(self):
                    a_var = 10
                    a_var += 2
        """)
        start = code.rindex("2")
        end = code.rindex("2") + 1
        refactored = self.do_extract_method(code, start, end, "new_func")
        expected = dedent("""\
            class A(object):
                def a_func(self):
                    a_var = 10
                    a_var += self.new_func()

                def new_func(self):
                    return 2
        """)
        self.assertEqual(expected, refactored)

    def test_breaks_and_continues_inside_loops(self):
        code = dedent("""\
            def a_func():
                for i in range(10):
                    continue
        """)
        start = code.index("for")
        end = len(code) - 1
        refactored = self.do_extract_method(code, start, end, "new_func")
        expected = dedent("""\
            def a_func():
                new_func()

            def new_func():
                for i in range(10):
                    continue
        """)
        self.assertEqual(expected, refactored)

    def test_breaks_and_continues_outside_loops(self):
        code = dedent("""\
            def a_func():
                for i in range(10):
                    a = i
                    continue
        """)
        start = code.index("a = i")
        end = len(code) - 1
        with self.assertRaises(rope.base.exceptions.RefactoringError):
            self.do_extract_method(code, start, end, "new_func")

    def test_for_loop_variable_scope(self):
        code = dedent("""\
            def my_func():
                i = 0
                for dummy in range(10):
                    i += 1
                    print(i)
        """)
        start, end = self._convert_line_range_to_offset(code, 4, 5)
        refactored = self.do_extract_method(code, start, end, "new_func")
        expected = dedent("""\
            def my_func():
                i = 0
                for dummy in range(10):
                    i = new_func(i)

            def new_func(i):
                i += 1
                print(i)
                return i
        """)
        self.assertEqual(expected, refactored)

    def test_for_loop_variable_scope_read_then_write(self):
        code = dedent("""\
            def my_func():
                i = 0
                for dummy in range(10):
                    a = i + 1
                    i = a + 1
        """)
        start, end = self._convert_line_range_to_offset(code, 4, 5)
        refactored = self.do_extract_method(code, start, end, "new_func")
        expected = dedent("""\
            def my_func():
                i = 0
                for dummy in range(10):
                    i = new_func(i)

            def new_func(i):
                a = i + 1
                i = a + 1
                return i
        """)
        self.assertEqual(expected, refactored)

    def test_for_loop_variable_scope_write_then_read(self):
        code = dedent("""\
            def my_func():
                i = 0
                for dummy in range(10):
                    i = 'hello'
                    print(i)
        """)
        start, end = self._convert_line_range_to_offset(code, 4, 5)
        refactored = self.do_extract_method(code, start, end, "new_func")
        expected = dedent("""\
            def my_func():
                i = 0
                for dummy in range(10):
                    new_func()

            def new_func():
                i = 'hello'
                print(i)
        """)
        self.assertEqual(expected, refactored)

    def test_for_loop_variable_scope_write_only(self):
        code = dedent("""\
            def my_func():
                i = 0
                for num in range(10):
                    i = 'hello' + num
                    print(i)
        """)
        start, end = self._convert_line_range_to_offset(code, 4, 4)
        refactored = self.do_extract_method(code, start, end, "new_func")
        expected = dedent("""\
            def my_func():
                i = 0
                for num in range(10):
                    i = new_func(num)
                    print(i)

            def new_func(num):
                i = 'hello' + num
                return i
        """)
        self.assertEqual(expected, refactored)

    def test_variable_writes_followed_by_variable_reads_after_extraction(self):
        code = dedent("""\
            def a_func():
                a = 1
                a = 2
                b = a
        """)
        start = code.index("a = 1")
        end = code.index("a = 2") - 1
        refactored = self.do_extract_method(code, start, end, "new_func")
        expected = dedent("""\
            def a_func():
                new_func()
                a = 2
                b = a

            def new_func():
                a = 1
        """)
        self.assertEqual(expected, refactored)

    def test_var_writes_followed_by_var_reads_inside_extraction(self):
        code = dedent("""\
            def a_func():
                a = 1
                a = 2
                b = a
        """)
        start = code.index("a = 2")
        end = len(code) - 1
        refactored = self.do_extract_method(code, start, end, "new_func")
        expected = dedent("""\
            def a_func():
                a = 1
                new_func()

            def new_func():
                a = 2
                b = a
        """)
        self.assertEqual(expected, refactored)

    def test_extract_variable(self):
        code = dedent("""\
            a_var = 10 + 20
        """)
        start = code.index("10")
        end = code.index("20") + 2
        refactored = self.do_extract_variable(code, start, end, "new_var")
        expected = dedent("""\
            new_var = 10 + 20
            a_var = new_var
        """)
        self.assertEqual(expected, refactored)

    @testutils.only_for_versions_higher("3.6")
    def test_extract_variable_f_string(self):
        code = dedent("""\
            foo(f"abc {a_var} def", 10)
        """)
        start = code.index('f"')
        end = code.index('def"') + 4
        refactored = self.do_extract_variable(code, start, end, "new_var")
        expected = dedent("""\
            new_var = f"abc {a_var} def"
            foo(new_var, 10)
        """)
        self.assertEqual(expected, refactored)

    def test_extract_variable_multiple_lines(self):
        code = dedent("""\
            a = 1
            b = 2
        """)
        start = code.index("1")
        end = code.index("1") + 1
        refactored = self.do_extract_variable(code, start, end, "c")
        expected = dedent("""\
            c = 1
            a = c
            b = 2
        """)
        self.assertEqual(expected, refactored)

    def test_extract_variable_in_the_middle_of_statements(self):
        code = dedent("""\
            a = 1 + 2
        """)
        start = code.index("1")
        end = code.index("1") + 1
        refactored = self.do_extract_variable(code, start, end, "c")
        expected = dedent("""\
            c = 1
            a = c + 2
        """)
        self.assertEqual(expected, refactored)

    def test_extract_variable_for_a_tuple(self):
        code = dedent("""\
            a = 1, 2
        """)
        start = code.index("1")
        end = code.index("2") + 1
        refactored = self.do_extract_variable(code, start, end, "c")
        expected = dedent("""\
            c = 1, 2
            a = c
        """)
        self.assertEqual(expected, refactored)

    def test_extract_variable_for_a_string(self):
        code = dedent("""\
            def a_func():
                a = "hey!"
        """)
        start = code.index('"')
        end = code.rindex('"') + 1
        refactored = self.do_extract_variable(code, start, end, "c")
        expected = dedent("""\
            def a_func():
                c = "hey!"
                a = c
        """)
        self.assertEqual(expected, refactored)

    def test_extract_variable_inside_ifs(self):
        code = dedent("""\
            if True:
                a = 1 + 2
        """)
        start = code.index("1")
        end = code.rindex("2") + 1
        refactored = self.do_extract_variable(code, start, end, "b")
        expected = dedent("""\
            if True:
                b = 1 + 2
                a = b
        """)
        self.assertEqual(expected, refactored)

    def test_extract_variable_inside_ifs_and_logical_lines(self):
        code = dedent("""\
            if True:
                a = (3 +
            (1 + 2))
        """)
        start = code.index("1")
        end = code.index("2") + 1
        refactored = self.do_extract_variable(code, start, end, "b")
        expected = dedent("""\
            if True:
                b = 1 + 2
                a = (3 +
            (b))
        """)
        self.assertEqual(expected, refactored)

    # TODO: Handle when extracting a subexpression
    def xxx_test_extract_variable_for_a_subexpression(self):
        code = dedent("""\
            a = 3 + 1 + 2
        """)
        start = code.index("1")
        end = code.index("2") + 1
        refactored = self.do_extract_variable(code, start, end, "b")
        expected = dedent("""\
            b = 1 + 2
            a = 3 + b
        """)
        self.assertEqual(expected, refactored)

    def test_extract_variable_starting_from_the_start_of_the_line(self):
        code = dedent("""\
            a_dict = {1: 1}
            a_dict.values().count(1)
        """)
        start = code.rindex("a_dict")
        end = code.index("count") - 1
        refactored = self.do_extract_variable(code, start, end, "values")
        expected = dedent("""\
            a_dict = {1: 1}
            values = a_dict.values()
            values.count(1)
        """)
        self.assertEqual(expected, refactored)

    def test_extract_variable_on_the_last_line_of_a_function(self):
        code = dedent("""\
            def f():
                a_var = {}
                a_var.keys()
        """)
        start = code.rindex("a_var")
        end = code.index(".keys")
        refactored = self.do_extract_variable(code, start, end, "new_var")
        expected = dedent("""\
            def f():
                a_var = {}
                new_var = a_var
                new_var.keys()
        """)
        self.assertEqual(expected, refactored)

    def test_extract_variable_on_the_indented_function_statement(self):
        code = dedent("""\
            def f():
                if True:
                    a_var = 1 + 2
        """)
        start = code.index("1")
        end = code.index("2") + 1
        refactored = self.do_extract_variable(code, start, end, "new_var")
        expected = dedent("""\
            def f():
                if True:
                    new_var = 1 + 2
                    a_var = new_var
        """)
        self.assertEqual(expected, refactored)

    def test_extract_method_on_the_last_line_of_a_function(self):
        code = dedent("""\
            def f():
                a_var = {}
                a_var.keys()
        """)
        start = code.rindex("a_var")
        end = code.index(".keys")
        refactored = self.do_extract_method(code, start, end, "new_f")
        expected = dedent("""\
            def f():
                a_var = {}
                new_f(a_var).keys()

            def new_f(a_var):
                return a_var
        """)
        self.assertEqual(expected, refactored)

    def test_raising_exception_when_on_incomplete_variables(self):
        code = dedent("""\
            a_var = 10 + 20
        """)
        start = code.index("10") + 1
        end = code.index("20") + 2
        with self.assertRaises(rope.base.exceptions.RefactoringError):
            self.do_extract_method(code, start, end, "new_func")

    def test_raising_exception_when_on_incomplete_variables_on_end(self):
        code = dedent("""\
            a_var = 10 + 20
        """)
        start = code.index("10")
        end = code.index("20") + 1
        with self.assertRaises(rope.base.exceptions.RefactoringError):
            self.do_extract_method(code, start, end, "new_func")

    def test_raising_exception_on_bad_parens(self):
        code = dedent("""\
            a_var = (10 + 20) + 30
        """)
        start = code.index("20")
        end = code.index("30") + 2
        with self.assertRaises(rope.base.exceptions.RefactoringError):
            self.do_extract_method(code, start, end, "new_func")

    def test_raising_exception_on_bad_operators(self):
        code = dedent("""\
            a_var = 10 + 20 + 30
        """)
        start = code.index("10")
        end = code.rindex("+") + 1
        with self.assertRaises(rope.base.exceptions.RefactoringError):
            self.do_extract_method(code, start, end, "new_func")

    # FIXME: Extract method should be more intelligent about bad ranges
    def xxx_test_raising_exception_on_function_parens(self):
        code = dedent("""\
            a = range(10)""")
        start = code.index("(")
        end = code.rindex(")") + 1
        with self.assertRaises(rope.base.exceptions.RefactoringError):
            self.do_extract_method(code, start, end, "new_func")
    
    def test_raising_exception_on_incomplete_block(self):
        code = dedent("""\
            if True:
                a = 1
                b = 2
        """)
        start = code.index("if")
        end = code.index("1") + 1
        with self.assertRaises(rope.base.exceptions.RefactoringError):
            self.do_extract_method(code, start, end, "new_func")

    def test_raising_exception_on_incomplete_block_2(self):
        code = dedent("""\
            if True:
                a = 1
            #
                b = 2
        """)
        start = code.index("if")
        end = code.index("1") + 1
        with self.assertRaises(rope.base.exceptions.RefactoringError):
            self.do_extract_method(code, start, end, "new_func")

    def test_raising_exception_on_incomplete_block_3(self):
        code = dedent("""\
            if True:
                a = 1
            
                b = 2
        """)
        start = code.index("if")
        end = code.index("1") + 1
        with self.assertRaises(rope.base.exceptions.RefactoringError):
            self.do_extract_method(code, start, end, "new_func")

    def test_raising_exception_on_incomplete_block_4(self):
        code = dedent("""\
                #
            if True:
                a = 1
                b = 2
        """)
        start = code.index("#")
        end = code.index("1") + 1
        with self.assertRaises(rope.base.exceptions.RefactoringError):
            self.do_extract_method(code, start, end, "new_func")

    def test_raising_exception_on_incomplete_block_5(self):
        code = dedent("""\
            if True:
                if 0:
                    a = 1
        """)
        start = code.index("if")
        end = code.index("0:") + 2
        with self.assertRaises(rope.base.exceptions.RefactoringError):
            self.do_extract_method(code, start, end, "new_func")

    def test_extract_method_and_extra_blank_lines(self):
        code = dedent("""\

            print(1)
        """)
        refactored = self.do_extract_method(code, 0, len(code), "new_f")
        expected = dedent("""\


            def new_f():
                print(1)

            new_f()
        """)
        self.assertEqual(expected, refactored)

    @testutils.only_for_versions_higher("3.6")
    def test_extract_method_f_string_extract_method(self):
        code = dedent("""\
            def func(a_var):
                foo(f"abc {a_var}", 10)
        """)
        start = code.index('f"')
        end = code.index('}"') + 2
        refactored = self.do_extract_method(code, start, end, "new_func")
        expected = dedent("""\
            def func(a_var):
                foo(new_func(a_var), 10)

            def new_func(a_var):
                return f"abc {a_var}"
        """)
        self.assertEqual(expected, refactored)

    @testutils.only_for_versions_higher("3.6")
    def test_extract_method_f_string_extract_method_complex_expression(self):
        code = dedent("""\
            def func(a_var):
                b_var = int
                c_var = 10
                fill = 10
                foo(f"abc {a_var + f'{b_var(a_var)}':{fill}16}" f"{c_var}", 10)
        """)
        start = code.index('f"')
        end = code.index('c_var}"') + 7
        refactored = self.do_extract_method(code, start, end, "new_func")
        expected = dedent("""\
            def func(a_var):
                b_var = int
                c_var = 10
                fill = 10
                foo(new_func(a_var, b_var, c_var, fill), 10)

            def new_func(a_var, b_var, c_var, fill):
                return f"abc {a_var + f'{b_var(a_var)}':{fill}16}" f"{c_var}"
        """)
        self.assertEqual(expected, refactored)

    @testutils.only_for_versions_higher("3.6")
    def test_extract_method_f_string_false_comment(self):
        code = dedent("""\
            def func(a_var):
                foo(f"abc {a_var} # ", 10)
        """)
        start = code.index('f"')
        end = code.index('# "') + 3
        refactored = self.do_extract_method(code, start, end, "new_func")
        expected = dedent("""\
            def func(a_var):
                foo(new_func(a_var), 10)

            def new_func(a_var):
                return f"abc {a_var} # "
        """)
        self.assertEqual(expected, refactored)

    @unittest.expectedFailure
    @testutils.only_for_versions_higher("3.6")
    def test_extract_method_f_string_false_format_value_in_regular_string(self):
        code = dedent("""\
            def func(a_var):
                b_var = 1
                foo(f"abc {a_var} " "{b_var}" f"{b_var} def", 10)
        """)
        start = code.index('f"')
        end = code.index('def"') + 4
        refactored = self.do_extract_method(code, start, end, "new_func")
        expected = dedent("""\
            def func(a_var):
                b_var = 1
                foo(new_func(a_var, b_var), 10)

            def new_func(a_var, b_var):
                return f"abc {a_var} " "{b_var}" f"{b_var} def"
        """)
        self.assertEqual(expected, refactored)

    def test_variable_writes_in_the_same_line_as_variable_read(self):
        code = dedent("""\
            a = 1
            a = 1 + a
        """)
        start = code.index("\n") + 1
        end = len(code)
        refactored = self.do_extract_method(code, start, end, "new_f", global_=True)
        expected = dedent("""\
            a = 1

            def new_f(a):
                a = 1 + a

            new_f(a)
        """)
        self.assertEqual(expected, refactored)

    def test_variable_writes_in_the_same_line_as_variable_read2(self):
        code = dedent("""\
            a = 1
            a += 1
        """)
        start = code.index("\n") + 1
        end = len(code)
        refactored = self.do_extract_method(code, start, end, "new_f", global_=True)
        expected = dedent("""\
            a = 1

            def new_f(a):
                a += 1

            new_f(a)
        """)
        self.assertEqual(expected, refactored)

    def test_variable_writes_in_the_same_line_as_variable_read3(self):
        code = dedent("""\
            a = 1
            a += 1
            print(a)
        """)
        start, end = self._convert_line_range_to_offset(code, 2, 2)
        refactored = self.do_extract_method(code, start, end, "new_f")
        expected = dedent("""\
            a = 1

            def new_f(a):
                a += 1
                return a

            a = new_f(a)
            print(a)
        """)
        self.assertEqual(expected, refactored)

    def test_variable_writes_only(self):
        code = dedent("""\
            i = 1
            print(i)
        """)
        start, end = self._convert_line_range_to_offset(code, 1, 1)
        refactored = self.do_extract_method(code, start, end, "new_f")
        expected = dedent("""\

            def new_f():
                i = 1
                return i

            i = new_f()
            print(i)
        """)
        self.assertEqual(expected, refactored)

    def test_variable_and_similar_expressions(self):
        code = dedent("""\
            a = 1
            b = 1
        """)
        start = code.index("1")
        end = start + 1
        refactored = self.do_extract_variable(code, start, end, "one", similar=True)
        expected = dedent("""\
            one = 1
            a = one
            b = one
        """)
        self.assertEqual(expected, refactored)

    def test_definition_should_appear_before_the_first_use(self):
        code = dedent("""\
            a = 1
            b = 1
        """)
        start = code.rindex("1")
        end = start + 1
        refactored = self.do_extract_variable(code, start, end, "one", similar=True)
        expected = dedent("""\
            one = 1
            a = one
            b = one
        """)
        self.assertEqual(expected, refactored)

    def test_extract_method_and_similar_expressions(self):
        code = dedent("""\
            a = 1
            b = 1
        """)
        start = code.index("1")
        end = start + 1
        refactored = self.do_extract_method(code, start, end, "one", similar=True)
        expected = dedent("""\

            def one():
                return 1

            a = one()
            b = one()
        """)
        self.assertEqual(expected, refactored)

    def test_simple_extract_method_and_similar_statements(self):
        code = dedent("""\
            class AClass(object):

                def func1(self):
                    a = 1 + 2
                    b = a
                def func2(self):
                    a = 1 + 2
                    b = a
        """)
        start, end = self._convert_line_range_to_offset(code, 4, 4)
        refactored = self.do_extract_method(code, start, end, "new_func", similar=True)
        expected = dedent("""\
            class AClass(object):

                def func1(self):
                    a = self.new_func()
                    b = a

                def new_func(self):
                    a = 1 + 2
                    return a
                def func2(self):
                    a = self.new_func()
                    b = a
        """)
        self.assertEqual(expected, refactored)

    def test_extract_method_and_similar_statements2(self):
        code = dedent("""\
            class AClass(object):

                def func1(self, p1):
                    a = p1 + 2
                def func2(self, p2):
                    a = p2 + 2
        """)
        start = code.rindex("p1")
        end = code.index("2\n") + 1
        refactored = self.do_extract_method(code, start, end, "new_func", similar=True)
        expected = dedent("""\
            class AClass(object):

                def func1(self, p1):
                    a = self.new_func(p1)

                def new_func(self, p1):
                    return p1 + 2
                def func2(self, p2):
                    a = self.new_func(p2)
        """)
        self.assertEqual(expected, refactored)

    def test_extract_method_and_similar_sttemnts_return_is_different(self):
        code = dedent("""\
            class AClass(object):

                def func1(self, p1):
                    a = p1 + 2
                def func2(self, p2):
                    self.attr = p2 + 2
        """)
        start = code.rindex("p1")
        end = code.index("2\n") + 1
        refactored = self.do_extract_method(code, start, end, "new_func", similar=True)
        expected = dedent("""\
            class AClass(object):

                def func1(self, p1):
                    a = self.new_func(p1)

                def new_func(self, p1):
                    return p1 + 2
                def func2(self, p2):
                    self.attr = self.new_func(p2)
        """)
        self.assertEqual(expected, refactored)

    def test_extract_method_and_similar_sttemnts_overlapping_regions(self):
        code = dedent("""\
            def func(p):
                a = p
                b = a
                c = b
                d = c
                return d""")
        start = code.index("a")
        end = code.rindex("a") + 1
        refactored = self.do_extract_method(code, start, end, "new_func", similar=True)
        expected = dedent("""\
            def func(p):
                b = new_func(p)
                d = new_func(b)
                return d
            def new_func(p):
                a = p
                b = a
                return b
        """)
        self.assertEqual(expected, refactored)

    def test_definition_should_appear_where_it_is_visible(self):
        code = dedent("""\
            if True:
                a = 1
            else:
                b = 1
        """)
        start = code.rindex("1")
        end = start + 1
        refactored = self.do_extract_variable(code, start, end, "one", similar=True)
        expected = dedent("""\
            one = 1
            if True:
                a = one
            else:
                b = one
        """)
        self.assertEqual(expected, refactored)

    def test_extract_variable_and_similar_statements_in_classes(self):
        code = dedent("""\
            class AClass(object):

                def func1(self):
                    a = 1
                def func2(self):
                    b = 1
        """)
        start = code.index(" 1") + 1
        refactored = self.do_extract_variable(
            code, start, start + 1, "one", similar=True
        )
        expected = dedent("""\
            class AClass(object):

                def func1(self):
                    one = 1
                    a = one
                def func2(self):
                    b = 1
        """)
        self.assertEqual(expected, refactored)

    def test_extract_method_in_staticmethods(self):
        code = dedent("""\
            class AClass(object):

                @staticmethod
                def func2():
                    b = 1
        """)
        start = code.index(" 1") + 1
        refactored = self.do_extract_method(code, start, start + 1, "one", similar=True)
        expected = dedent("""\
            class AClass(object):

                @staticmethod
                def func2():
                    b = AClass.one()

                @staticmethod
                def one():
                    return 1
        """)
        self.assertEqual(expected, refactored)

    def test_extract_normal_method_with_staticmethods(self):
        code = dedent("""\
            class AClass(object):

                @staticmethod
                def func1():
                    b = 1
                def func2(self):
                    b = 1
        """)
        start = code.rindex(" 1") + 1
        refactored = self.do_extract_method(code, start, start + 1, "one", similar=True)
        expected = dedent("""\
            class AClass(object):

                @staticmethod
                def func1():
                    b = 1
                def func2(self):
                    b = self.one()

                def one(self):
                    return 1
        """)
        self.assertEqual(expected, refactored)

    def test_extract_variable_with_no_new_lines_at_the_end(self):
        code = "a_var = 10"
        start = code.index("10")
        end = start + 2
        refactored = self.do_extract_variable(code, start, end, "new_var")
        expected = dedent("""\
            new_var = 10
            a_var = new_var""")
        self.assertEqual(expected, refactored)

    def test_extract_method_containing_return_in_functions(self):
        code = dedent("""\
            def f(arg):
                return arg
            print(f(1))
        """)
        start, end = self._convert_line_range_to_offset(code, 1, 3)
        refactored = self.do_extract_method(code, start, end, "a_func")
        expected = dedent("""\

            def a_func():
                def f(arg):
                    return arg
                print(f(1))

            a_func()
        """)
        self.assertEqual(expected, refactored)

    def test_extract_method_and_varying_first_parameter(self):
        code = dedent("""\
            class C(object):
                def f1(self):
                    print(str(self))
                def f2(self):
                    print(str(1))
        """)
        start = code.index("print(") + 6
        end = code.index("))\n") + 1
        refactored = self.do_extract_method(code, start, end, "to_str", similar=True)
        expected = dedent("""\
            class C(object):
                def f1(self):
                    print(self.to_str())

                def to_str(self):
                    return str(self)
                def f2(self):
                    print(str(1))
        """)
        self.assertEqual(expected, refactored)

    def test_extract_method_when_an_attribute_exists_in_function_scope(self):
        code = dedent("""\
            class A(object):
                def func(self):
                    pass
            a = A()
            def f():
                func = a.func()
                print(func)
        """)

        start, end = self._convert_line_range_to_offset(code, 6, 6)
        refactored = self.do_extract_method(code, start, end, "g")
        refactored = refactored[refactored.index("A()") + 4 :]
        expected = dedent("""\
            def f():
                func = g()
                print(func)

            def g():
                func = a.func()
                return func
        """)
        self.assertEqual(expected, refactored)

    def test_global_option_for_extract_method(self):
        code = dedent("""\
            def a_func():
                print(1)
        """)
        start, end = self._convert_line_range_to_offset(code, 2, 2)
        refactored = self.do_extract_method(code, start, end, "extracted", global_=True)
        expected = dedent("""\
            def a_func():
                extracted()

            def extracted():
                print(1)
        """)
        self.assertEqual(expected, refactored)

    def test_global_extract_method(self):
        code = dedent("""\
            class AClass(object):

                def a_func(self):
                    print(1)
        """)
        start, end = self._convert_line_range_to_offset(code, 4, 4)
        refactored = self.do_extract_method(code, start, end, "new_func", global_=True)
        expected = dedent("""\
            class AClass(object):

                def a_func(self):
                    new_func()

            def new_func():
                print(1)
        """)
        self.assertEqual(expected, refactored)

    def test_global_extract_method_with_multiple_methods(self):
        code = dedent("""\
            class AClass(object):
                def a_func(self):
                    print(1)

                def another_func(self):
                    pass
        """)
        start, end = self._convert_line_range_to_offset(code, 3, 3)
        refactored = self.do_extract_method(code, start, end, "new_func", global_=True)
        expected = dedent("""\
            class AClass(object):
                def a_func(self):
                    new_func()

                def another_func(self):
                    pass

            def new_func():
                print(1)
        """)
        self.assertEqual(expected, refactored)

    def test_where_to_search_when_extracting_global_names(self):
        code = dedent("""\
            def a():
                return 1
            def b():
                return 1
            b = 1
        """)
        start = code.index("1")
        end = start + 1
        refactored = self.do_extract_variable(
            code, start, end, "one", similar=True, global_=True
        )
        expected = dedent("""\
            def a():
                return one
            one = 1
            def b():
                return one
            b = one
        """)
        self.assertEqual(expected, refactored)

    def test_extracting_pieces_with_distinct_temp_names(self):
        code = dedent("""\
            a = 1
            print(a)
            b = 1
            print(b)
        """)
        start = code.index("a")
        end = code.index("\nb")
        refactored = self.do_extract_method(
            code, start, end, "f", similar=True, global_=True
        )
        expected = dedent("""\

            def f():
                a = 1
                print(a)

            f()
            f()
        """)
        self.assertEqual(expected, refactored)

    def test_extract_methods_in_glob_funcs_should_be_glob(self):
        code = dedent("""\
            def f():
                a = 1
            def g():
                b = 1
        """)
        start = code.rindex("1")
        refactored = self.do_extract_method(
            code, start, start + 1, "one", similar=True, global_=False
        )
        expected = dedent("""\
            def f():
                a = one()
            def g():
                b = one()

            def one():
                return 1
        """)
        self.assertEqual(expected, refactored)

    def test_extract_methods_in_glob_funcs_should_be_glob_2(self):
        code = dedent("""\
            if 1:
                var = 2
        """)
        start = code.rindex("2")
        refactored = self.do_extract_method(
            code, start, start + 1, "two", similar=True, global_=False
        )
        expected = dedent("""\

            def two():
                return 2

            if 1:
                var = two()
        """)
        self.assertEqual(expected, refactored)

    def test_extract_method_and_try_blocks(self):
        code = dedent("""\
            def f():
                try:
                    pass
                except Exception:
                    pass
        """)
        start, end = self._convert_line_range_to_offset(code, 2, 5)
        refactored = self.do_extract_method(code, start, end, "g")
        expected = dedent("""\
            def f():
                g()

            def g():
                try:
                    pass
                except Exception:
                    pass
        """)
        self.assertEqual(expected, refactored)

    @testutils.only_for_versions_higher("3.11")
    def test_extract_method_and_try_except_star_block_1(self):
        code = dedent("""\
            def f():
                try:
                    pass
                except* Exception:
                    pass
        """)
        start, end = self._convert_line_range_to_offset(code, 2, 5)
        refactored = self.do_extract_method(code, start, end, "g")
        expected = dedent("""\
            def f():
                g()

            def g():
                try:
                    pass
                except* Exception:
                    pass
        """)
        self.assertEqual(expected, refactored)

    def test_extract_method_and_augmented_assignment_nested_1(self):
        code = dedent("""\
            def f():
                my_var = [[0], [1], [2]]
                my_var[0][0] += 1
                print(1)
        """)
        start, end = self._convert_line_range_to_offset(code, 4, 4)
        refactored = self.do_extract_method(code, start, end, "g")
        expected = dedent("""\
            def f():
                my_var = [[0], [1], [2]]
                my_var[0][0] += 1
                g()

            def g():
                print(1)
        """)
        self.assertEqual(expected, refactored)

    def test_extract_method_and_augmented_assignment_nested_2(self):
        code = dedent("""\
            def f():
                my_var = [[0], [1], [2]]
                my_var[0][0] += 1
                print(my_var)
        """)
        start, end = self._convert_line_range_to_offset(code, 3, 3)
        refactored = self.do_extract_method(code, start, end, "g")
        expected = dedent("""\
            def f():
                my_var = [[0], [1], [2]]
                g(my_var)
                print(my_var)

            def g(my_var):
                my_var[0][0] += 1
        """)
        self.assertEqual(expected, refactored)

    def test_extract_method_and_augmented_assignment_var_to_read_in_lhs(self):
        code = dedent("""\
            def f():
                var_to_read = 0
                my_var = [0, 1, 2]
                my_var[var_to_read] += 1
                print(my_var)
        """)
        start, end = self._convert_line_range_to_offset(code, 4, 4)
        refactored = self.do_extract_method(code, start, end, "g")
        expected = dedent("""\
            def f():
                var_to_read = 0
                my_var = [0, 1, 2]
                g(my_var, var_to_read)
                print(my_var)

            def g(my_var, var_to_read):
                my_var[var_to_read] += 1
        """)
        self.assertEqual(expected, refactored)

    def test_extract_method_and_augmented_assignment_in_try_block(self):
        code = dedent("""\
            def f():
                any_subscriptable = [0]
                try:
                    any_subscriptable[0] += 1
                except Exception:
                    pass
        """)
        start, end = self._convert_line_range_to_offset(code, 2, 6)
        refactored = self.do_extract_method(code, start, end, "g")
        expected = dedent("""\
            def f():
                g()

            def g():
                any_subscriptable = [0]
                try:
                    any_subscriptable[0] += 1
                except Exception:
                    pass
        """)
        self.assertEqual(expected, refactored)

    def test_extract_and_not_passing_global_functions(self):
        code = dedent("""\
            def next(p):
                return p + 1
            var = next(1)
        """)
        start = code.rindex("next")
        refactored = self.do_extract_method(code, start, len(code) - 1, "two")
        expected = dedent("""\
            def next(p):
                return p + 1

            def two():
                return next(1)

            var = two()
        """)
        self.assertEqual(expected, refactored)

    def test_extracting_with_only_one_return(self):
        code = dedent("""\
            def f():
                var = 1
                return var
        """)
        start, end = self._convert_line_range_to_offset(code, 2, 3)
        refactored = self.do_extract_method(code, start, end, "g")
        expected = dedent("""\
            def f():
                return g()

            def g():
                var = 1
                return var
        """)
        self.assertEqual(expected, refactored)

    def test_extracting_variable_and_implicit_continuations(self):
        code = dedent("""\
            s = ("1"
              "2")
        """)
        start = code.index('"')
        end = code.rindex('"') + 1
        refactored = self.do_extract_variable(code, start, end, "s2")
        expected = dedent("""\
            s2 = "1" "2"
            s = (s2)
        """)
        self.assertEqual(expected, refactored)

    def test_extracting_method_and_implicit_continuations(self):
        code = dedent("""\
            s = ("1"
              "2")
        """)
        start = code.index('"')
        end = code.rindex('"') + 1
        refactored = self.do_extract_method(code, start, end, "f")
        expected = dedent("""\

            def f():
                return "1" "2"

            s = (f())
        """)
        self.assertEqual(expected, refactored)

    def test_passing_conditional_updated_vars_in_extracted(self):
        code = dedent("""\
            def f(a):
                if 0:
                    a = 1
                print(a)
        """)
        start, end = self._convert_line_range_to_offset(code, 2, 4)
        refactored = self.do_extract_method(code, start, end, "g")
        expected = dedent("""\
            def f(a):
                g(a)

            def g(a):
                if 0:
                    a = 1
                print(a)
        """)
        self.assertEqual(expected, refactored)

    def test_returning_conditional_updated_vars_in_extracted(self):
        code = dedent("""\
            def f(a):
                if 0:
                    a = 1
                print(a)
        """)
        start, end = self._convert_line_range_to_offset(code, 2, 3)
        refactored = self.do_extract_method(code, start, end, "g")
        expected = dedent("""\
            def f(a):
                a = g(a)
                print(a)

            def g(a):
                if 0:
                    a = 1
                return a
        """)
        self.assertEqual(expected, refactored)

    def test_extract_method_with_variables_possibly_written_to(self):
        code = dedent("""\
            def a_func(b):
                if b > 0:
                    a = 2
                print(a)
        """)
        start, end = self._convert_line_range_to_offset(code, 2, 3)
        refactored = self.do_extract_method(code, start, end, "extracted")
        expected = dedent("""\
            def a_func(b):
                a = extracted(b)
                print(a)

            def extracted(b):
                if b > 0:
                    a = 2
                return a
        """)
        self.assertEqual(expected, refactored)

    def test_extract_method_with_list_comprehension(self):
        code = dedent("""\
            def foo():
                x = [e for e in []]
                f = 23

                for e, f in []:
                    def bar():
                        e[42] = 1
        """)
        start, end = self._convert_line_range_to_offset(code, 4, 7)
        refactored = self.do_extract_method(code, start, end, "baz")
        expected = dedent("""\
            def foo():
                x = [e for e in []]
                f = 23

                baz()

            def baz():
                for e, f in []:
                    def bar():
                        e[42] = 1
        """)
        self.assertEqual(expected, refactored)

    def test_extract_method_with_list_comprehension_in_class_method(self):
        code = dedent("""\
            class SomeClass:
                def method(self):
                    result = [i for i in range(1)]
                    print(1)
        """)
        start, end = self._convert_line_range_to_offset(code, 4, 4)
        refactored = self.do_extract_method(code, start, end, "baz", similar=True)
        expected = dedent("""\
            class SomeClass:
                def method(self):
                    result = [i for i in range(1)]
                    self.baz()

                def baz(self):
                    print(1)
        """)
        self.assertEqual(expected, refactored)

    def test_extract_method_with_list_comprehension_and_iter(self):
        code = dedent("""\
            def foo():
                x = [e for e in []]
                f = 23

                for x, f in x:
                    def bar():
                        x[42] = 1
        """)
        start, end = self._convert_line_range_to_offset(code, 4, 7)
        refactored = self.do_extract_method(code, start, end, "baz")
        expected = dedent("""\
            def foo():
                x = [e for e in []]
                f = 23

                baz(x)

            def baz(x):
                for x, f in x:
                    def bar():
                        x[42] = 1
        """)
        self.assertEqual(expected, refactored)

    def test_extract_method_with_list_comprehension_and_orelse(self):
        code = dedent("""\
            def foo():
                x = [e for e in []]
                f = 23

                for e, f in []:
                    def bar():
                        e[42] = 1
        """)
        start, end = self._convert_line_range_to_offset(code, 4, 7)
        refactored = self.do_extract_method(code, start, end, "baz")
        expected = dedent("""\
            def foo():
                x = [e for e in []]
                f = 23

                baz()

            def baz():
                for e, f in []:
                    def bar():
                        e[42] = 1
        """)
        self.assertEqual(expected, refactored)

    def test_extract_method_with_list_comprehension_multiple_targets(self):
        code = dedent("""\
            def foo():
                x = [(a, b) for a, b in []]
                f = 23
                print("hello")
        """)
        start, end = self._convert_line_range_to_offset(code, 4, 4)
        refactored = self.do_extract_method(code, start, end, "baz")
        expected = dedent("""\
            def foo():
                x = [(a, b) for a, b in []]
                f = 23
                baz()

            def baz():
                print("hello")
        """)
        self.assertEqual(expected, refactored)

    def test_extract_function_with_for_else_statemant(self):
        code = dedent("""\
            def a_func():
                for i in range(10):
                    a = i
                else:
                    a = None
        """)
        start = code.index("for")
        end = len(code) - 1
        refactored = self.do_extract_method(code, start, end, "new_func")
        expected = dedent("""\
            def a_func():
                new_func()

            def new_func():
                for i in range(10):
                    a = i
                else:
                    a = None
        """)
        self.assertEqual(expected, refactored)

    def test_extract_function_with_for_else_statemant_more(self):
        """TODO: fixed code to test passed"""
        code = dedent("""\
            def a_func():
                for i in range(10):
                    a = i
                else:
                    for i in range(5):
                        b = i
                    else:
                        b = None
                a = None
        """)

        start = code.index("for")
        end = len(code) - 1
        refactored = self.do_extract_method(code, start, end, "new_func")
        expected = dedent("""\
            def a_func():
                new_func()

            def new_func():
                for i in range(10):
                    a = i
                else:
                    for i in range(5):
                        b = i
                    else:
                        b = None
                a = None
        """)
        self.assertEqual(expected, refactored)

    def test_extract_function_with_for_else_statemant_outside_loops(self):
        code = dedent("""\
            def a_func():
                for i in range(10):
                    a = i
                else:
                    a=None
        """)
        start = code.index("a = i")
        end = len(code) - 1
        with self.assertRaises(rope.base.exceptions.RefactoringError):
            self.do_extract_method(code, start, end, "new_func")

    def test_extract_function_with_inline_assignment_in_method(self):
        code = dedent("""\
            def foo():
                i = 1
                i += 1
                print(i)
        """)
        start, end = self._convert_line_range_to_offset(code, 3, 3)
        refactored = self.do_extract_method(code, start, end, "new_func")
        expected = dedent("""\
            def foo():
                i = 1
                i = new_func(i)
                print(i)

            def new_func(i):
                i += 1
                return i
        """)
        self.assertEqual(expected, refactored)

    @testutils.only_for_versions_higher("3.8")
    def test_extract_function_statement_with_inline_assignment_in_condition(self):
        code = dedent("""\
            def foo(a):
                if i := a == 5:
                    i += 1
                print(i)
        """)
        start, end = self._convert_line_range_to_offset(code, 2, 3)
        refactored = self.do_extract_method(code, start, end, "new_func")
        expected = dedent("""\
            def foo(a):
                i = new_func(a)
                print(i)

            def new_func(a):
                if i := a == 5:
                    i += 1
                return i
        """)
        self.assertEqual(expected, refactored)

    @testutils.only_for_versions_higher("3.8")
    def test_extract_function_expression_with_inline_assignment_in_condition(self):
        code = dedent("""\
            def foo(a):
                if i := a == 5:
                    i += 1
                print(i)
        """)
        extract_target = "i := a == 5"
        start, end = code.index(extract_target), code.index(extract_target) + len(
            extract_target
        )
        refactored = self.do_extract_method(code, start, end, "new_func")
        expected = dedent("""\
            def foo(a):
                if i := new_func(a):
                    i += 1
                print(i)

            def new_func(a):
                return (i := a == 5)
        """)
        self.assertEqual(expected, refactored)

    @testutils.only_for_versions_higher("3.8")
    def test_extract_function_expression_with_inline_assignment_complex(self):
        code = dedent("""\
            def foo(a):
                if i := a == (c := 5):
                    i += 1
                    c += 1
                print(i)
        """)
        extract_target = "i := a == (c := 5)"
        start, end = code.index(extract_target), code.index(extract_target) + len(
            extract_target
        )
        refactored = self.do_extract_method(code, start, end, "new_func")
        expected = dedent("""\
            def foo(a):
                if i, c := new_func(a):
                    i += 1
                    c += 1
                print(i)

            def new_func(a):
                return (i := a == (c := 5))
        """)
        self.assertEqual(expected, refactored)

    @testutils.only_for_versions_higher("3.8")
    def test_extract_function_expression_with_inline_assignment_in_inner_expression(
        self,
    ):
        code = dedent("""\
            def foo(a):
                if a == (c := 5):
                    c += 1
                print(i)
        """)
        extract_target = "a == (c := 5)"
        start, end = code.index(extract_target), code.index(extract_target) + len(
            extract_target
        )
        with self.assertRaisesRegex(
            rope.base.exceptions.RefactoringError,
            "Extracted piece cannot contain named expression \\(:= operator\\).",
        ):
            self.do_extract_method(code, start, end, "new_func")

    def test_extract_exec(self):
        code = dedent("""\
            exec("def f(): pass", {})
        """)
        start, end = self._convert_line_range_to_offset(code, 1, 1)
        refactored = self.do_extract_method(code, start, end, "new_func")
        expected = dedent("""\

            def new_func():
                exec("def f(): pass", {})

            new_func()
        """)
        self.assertEqual(expected, refactored)

    @testutils.only_for_versions_higher("3.5")
    def test_extract_async_function(self):
        code = dedent("""\
            async def my_func(my_list):
                for x in my_list:
                    var = x + 1
                return var
        """)
        start, end = self._convert_line_range_to_offset(code, 3, 3)
        refactored = self.do_extract_method(code, start, end, "new_func")
        expected = dedent("""\
            async def my_func(my_list):
                for x in my_list:
                    var = new_func(x)
                return var

            def new_func(x):
                var = x + 1
                return var
        """)
        self.assertEqual(expected, refactored)

    @testutils.only_for_versions_higher("3.5")
    def test_extract_inner_async_function(self):
        code = dedent("""\
            def my_func(my_list):
                async def inner_func(my_list):
                    for x in my_list:
                        var = x + 1
                return inner_func
        """)
        start, end = self._convert_line_range_to_offset(code, 2, 4)
        refactored = self.do_extract_method(code, start, end, "new_func")
        expected = dedent("""\
            def my_func(my_list):
                inner_func = new_func(my_list)
                return inner_func

            def new_func(my_list):
                async def inner_func(my_list):
                    for x in my_list:
                        var = x + 1
                return inner_func
        """)
        self.assertEqual(expected, refactored)

    @testutils.only_for_versions_higher("3.5")
    def test_extract_around_inner_async_function(self):
        code = dedent("""\
            def my_func(lst):
                async def inner_func(obj):
                    for x in obj:
                        var = x + 1
                return map(inner_func, lst)
        """)
        start, end = self._convert_line_range_to_offset(code, 5, 5)
        refactored = self.do_extract_method(code, start, end, "new_func")
        expected = dedent("""\
            def my_func(lst):
                async def inner_func(obj):
                    for x in obj:
                        var = x + 1
                return new_func(inner_func, lst)

            def new_func(inner_func, lst):
                return map(inner_func, lst)
        """)
        self.assertEqual(expected, refactored)

    @testutils.only_for_versions_higher("3.5")
    def test_extract_refactor_around_async_for_loop(self):
        code = dedent("""\
            async def my_func(my_list):
                async for x in my_list:
                    var = x + 1
                return var
        """)
        start, end = self._convert_line_range_to_offset(code, 3, 3)
        refactored = self.do_extract_method(code, start, end, "new_func")
        expected = dedent("""\
            async def my_func(my_list):
                async for x in my_list:
                    var = new_func(x)
                return var

            def new_func(x):
                var = x + 1
                return var
        """)
        self.assertEqual(expected, refactored)

    @testutils.only_for_versions_higher("3.5")
    @testutils.only_for_versions_lower("3.8")
    def test_extract_refactor_containing_async_for_loop_should_error_before_py38(self):
        """
        Refactoring async/await syntaxes is only supported in Python 3.8 and
        higher because support for ast.PyCF_ALLOW_TOP_LEVEL_AWAIT was only
        added to the standard library in Python 3.8.
        """
        code = dedent("""\
            async def my_func(my_list):
                async for x in my_list:
                    var = x + 1
                return var
        """)
        start, end = self._convert_line_range_to_offset(code, 2, 3)
        with self.assertRaisesRegex(
            rope.base.exceptions.RefactoringError,
            "Extracted piece can only have async/await statements if Rope is running on Python 3.8 or higher",
        ):
            self.do_extract_method(code, start, end, "new_func")

    @testutils.only_for_versions_higher("3.8")
    def test_extract_refactor_containing_async_for_loop_is_supported_after_py38(self):
        code = dedent("""\
            async def my_func(my_list):
                async for x in my_list:
                    var = x + 1
                return var
        """)
        start, end = self._convert_line_range_to_offset(code, 2, 3)
        refactored = self.do_extract_method(code, start, end, "new_func")
        expected = dedent("""\
            async def my_func(my_list):
                var = new_func(my_list)
                return var

            def new_func(my_list):
                async for x in my_list:
                    var = x + 1
                return var
        """)
        self.assertEqual(expected, refactored)

    @testutils.only_for_versions_higher("3.5")
    def test_extract_await_expression(self):
        code = dedent("""\
            async def my_func(my_list):
                for url in my_list:
                    resp = await request(url)
                return resp
        """)
        selected = "request(url)"
        start, end = code.index(selected), code.index(selected) + len(selected)
        refactored = self.do_extract_method(code, start, end, "new_func")
        expected = dedent("""\
            async def my_func(my_list):
                for url in my_list:
                    resp = await new_func(url)
                return resp

            def new_func(url):
                return request(url)
        """)
        self.assertEqual(expected, refactored)

    def test_extract_to_staticmethod(self):
        code = dedent("""\
            class A:
                def first_method(self):
                    a_var = 1
                    b_var = a_var + 1
        """)
        extract_target = "a_var + 1"
        start, end = code.index(extract_target), code.index(extract_target) + len(
            extract_target
        )
        refactored = self.do_extract_method(
            code, start, end, "second_method", kind="staticmethod"
        )
        expected = dedent("""\
            class A:
                def first_method(self):
                    a_var = 1
                    b_var = A.second_method(a_var)

                @staticmethod
                def second_method(a_var):
                    return a_var + 1
        """)
        self.assertEqual(expected, refactored)

    def test_extract_to_staticmethod_when_self_in_body(self):
        code = dedent("""\
            class A:
                def first_method(self):
                    a_var = 1
                    b_var = self.a_var + 1
        """)
        extract_target = "self.a_var + 1"
        start, end = code.index(extract_target), code.index(extract_target) + len(
            extract_target
        )
        refactored = self.do_extract_method(
            code, start, end, "second_method", kind="staticmethod"
        )
        expected = dedent("""\
            class A:
                def first_method(self):
                    a_var = 1
                    b_var = A.second_method(self)

                @staticmethod
                def second_method(self):
                    return self.a_var + 1
        """)
        self.assertEqual(expected, refactored)

    def test_extract_from_function_to_staticmethod_raises_exception(self):
        code = dedent("""\
            def first_method():
                a_var = 1
                b_var = a_var + 1
        """)
        extract_target = "a_var + 1"
        start, end = code.index(extract_target), code.index(extract_target) + len(
            extract_target
        )
        with self.assertRaisesRegex(
            rope.base.exceptions.RefactoringError,
            "Cannot extract to staticmethod/classmethod outside class",
        ):
            self.do_extract_method(
                code, start, end, "second_method", kind="staticmethod"
            )

    def test_extract_method_in_classmethods(self):
        code = dedent("""\
            class AClass(object):
                @classmethod
                def func2(cls):
                    b = 1
        """)
        start = code.index(" 1") + 1
        refactored = self.do_extract_method(code, start, start + 1, "one", similar=True)
        expected = dedent("""\
            class AClass(object):
                @classmethod
                def func2(cls):
                    b = AClass.one()

                @classmethod
                def one(cls):
                    return 1
        """)
        self.assertEqual(expected, refactored)

    def test_extract_from_function_to_classmethod_raises_exception(self):
        code = dedent("""\
            def first_method():
                a_var = 1
                b_var = a_var + 1
        """)
        extract_target = "a_var + 1"
        start, end = code.index(extract_target), code.index(extract_target) + len(
            extract_target
        )
        with self.assertRaisesRegex(
            rope.base.exceptions.RefactoringError,
            "Cannot extract to staticmethod/classmethod outside class",
        ):
            self.do_extract_method(
                code, start, end, "second_method", kind="classmethod"
            )

    def test_extract_to_classmethod_when_self_in_body(self):
        code = dedent("""\
            class A:
                def first_method(self):
                    a_var = 1
                    b_var = self.a_var + 1
        """)
        extract_target = "self.a_var + 1"
        start, end = code.index(extract_target), code.index(extract_target) + len(
            extract_target
        )
        refactored = self.do_extract_method(
            code, start, end, "second_method", kind="classmethod"
        )
        expected = dedent("""\
            class A:
                def first_method(self):
                    a_var = 1
                    b_var = A.second_method(self)

                @classmethod
                def second_method(cls, self):
                    return self.a_var + 1
        """)
        self.assertEqual(expected, refactored)

    def test_extract_to_classmethod(self):
        code = dedent("""\
            class A:
                def first_method(self):
                    a_var = 1
                    b_var = a_var + 1
        """)
        extract_target = "a_var + 1"
        start, end = code.index(extract_target), code.index(extract_target) + len(
            extract_target
        )
        refactored = self.do_extract_method(
            code, start, end, "second_method", kind="classmethod"
        )
        expected = dedent("""\
            class A:
                def first_method(self):
                    a_var = 1
                    b_var = A.second_method(a_var)

                @classmethod
                def second_method(cls, a_var):
                    return a_var + 1
        """)
        self.assertEqual(expected, refactored)

    def test_extract_to_classmethod_when_name_starts_with_at_sign(self):
        code = dedent("""\
            class A:
                def first_method(self):
                    a_var = 1
                    b_var = a_var + 1
        """)
        extract_target = "a_var + 1"
        start, end = code.index(extract_target), code.index(extract_target) + len(
            extract_target
        )
        refactored = self.do_extract_method(code, start, end, "@second_method")
        expected = dedent("""\
            class A:
                def first_method(self):
                    a_var = 1
                    b_var = A.second_method(a_var)

                @classmethod
                def second_method(cls, a_var):
                    return a_var + 1
        """)
        self.assertEqual(expected, refactored)

    def test_extract_to_staticmethod_when_name_starts_with_dollar_sign(self):
        code = dedent("""\
            class A:
                def first_method(self):
                    a_var = 1
                    b_var = a_var + 1
        """)
        extract_target = "a_var + 1"
        start, end = code.index(extract_target), code.index(extract_target) + len(
            extract_target
        )
        refactored = self.do_extract_method(code, start, end, "$second_method")
        expected = dedent("""\
            class A:
                def first_method(self):
                    a_var = 1
                    b_var = A.second_method(a_var)

                @staticmethod
                def second_method(a_var):
                    return a_var + 1
        """)
        self.assertEqual(expected, refactored)

    def test_raises_exception_when_sign_in_name_and_kind_mismatch(self):
        with self.assertRaisesRegex(
            rope.base.exceptions.RefactoringError, "Kind and shortcut in name mismatch"
        ):
            self.do_extract_method("code", 0, 1, "$second_method", kind="classmethod")

    def test_extracting_from_static_with_function_arg(self):
        code = dedent("""\
                class A:
                    @staticmethod
                    def first_method(someargs):
                        b_var = someargs + 1
            """)

        extract_target = "someargs + 1"
        start, end = code.index(extract_target), code.index(extract_target) + len(
            extract_target
        )
        refactored = self.do_extract_method(code, start, end, "second_method")
        expected = dedent("""\
                class A:
                    @staticmethod
                    def first_method(someargs):
                        b_var = A.second_method(someargs)

                    @staticmethod
                    def second_method(someargs):
                        return someargs + 1
            """)

        self.assertEqual(expected, refactored)

    def test_extract_with_list_comprehension(self):
        code = dedent("""\
            def f():
                y = [1,2,3,4]
                a = sum([x for x in y])
                b = sum([x for x in y])

                print(a, b)

            f()
        """)
        extract_target = "    a = sum([x for x in y])\n"
        start, end = code.index(extract_target), code.index(extract_target) + len(
            extract_target
        )
        refactored = self.do_extract_method(code, start, end, "_a")
        expected = dedent("""\
            def f():
                y = [1,2,3,4]
                a = _a(y)
                b = sum([x for x in y])

                print(a, b)

            def _a(y):
                a = sum([x for x in y])
                return a

            f()
        """)
        self.assertEqual(expected, refactored)

    def test_extract_with_generator(self):
        code = dedent("""\
            def f():
                y = [1,2,3,4]
                a = sum(x for x in y)
                b = sum(x for x in y)

                print(a, b)

            f()
        """)
        extract_target = "    a = sum(x for x in y)\n"
        start, end = code.index(extract_target), code.index(extract_target) + len(
            extract_target
        )
        refactored = self.do_extract_method(code, start, end, "_a")
        expected = dedent("""\
            def f():
                y = [1,2,3,4]
                a = _a(y)
                b = sum(x for x in y)

                print(a, b)

            def _a(y):
                a = sum(x for x in y)
                return a

            f()
        """)
        self.assertEqual(expected, refactored)

    def test_extract_with_generator_2(self):
        code = dedent("""\
            def f():
                y = [1,2,3,4]
                a = sum(x for x in y)
        """)
        extract_target = "x for x in y"
        start, end = code.index(extract_target), code.index(extract_target) + len(
            extract_target
        )
        refactored = self.do_extract_method(code, start, end, "_a")
        expected = dedent("""\
            def f():
                y = [1,2,3,4]
                a = sum(_a(y))

            def _a(y):
                return (x for x in y)
        """)
        self.assertEqual(expected, refactored)

    def test_extract_with_set_comprehension(self):
        code = dedent("""\
            def f():
                y = [1,2,3,4]
                a = sum({x for x in y})
                b = sum({x for x in y})

                print(a, b)

            f()
        """)
        extract_target = "    a = sum({x for x in y})\n"
        start, end = code.index(extract_target), code.index(extract_target) + len(
            extract_target
        )
        refactored = self.do_extract_method(code, start, end, "_a")
        expected = dedent("""\
            def f():
                y = [1,2,3,4]
                a = _a(y)
                b = sum({x for x in y})

                print(a, b)

            def _a(y):
                a = sum({x for x in y})
                return a

            f()
        """)
        self.assertEqual(expected, refactored)

    def test_extract_with_dict_comprehension(self):
        code = dedent("""\
            def f():
                y = [1,2,3,4]
                a = sum({x: x for x in y})
                b = sum({x: x for x in y})

                print(a, b)

            f()
        """)
        extract_target = "    a = sum({x: x for x in y})\n"
        start, end = code.index(extract_target), code.index(extract_target) + len(
            extract_target
        )
        refactored = self.do_extract_method(code, start, end, "_a")
        expected = dedent("""\
            def f():
                y = [1,2,3,4]
                a = _a(y)
                b = sum({x: x for x in y})

                print(a, b)

            def _a(y):
                a = sum({x: x for x in y})
                return a

            f()
        """)
        self.assertEqual(expected, refactored)

    def test_extract_function_expression_with_assignment_to_attribute(self):
        code = dedent("""\
            class A(object):
                def func(self):
                    self.var_a = 1
                    var_bb = self.var_a
        """)
        extract_target = "= self.var_a"
        start, end = (
            code.index(extract_target) + 2,
            code.index(extract_target) + 2 + len(extract_target) - 2,
        )
        refactored = self.do_extract_method(code, start, end, "new_func", similar=True)
        expected = dedent("""\
            class A(object):
                def func(self):
                    self.var_a = 1
                    var_bb = self.new_func()

                def new_func(self):
                    return self.var_a
        """)

        self.assertEqual(expected, refactored)

    def test_extract_function_expression_with_assignment_index(self):
        code = dedent("""\
            class A(object):
                def func(self, val):
                    self[val] = 1
                    var_bb = self[val]
        """)
        extract_target = "= self[val]"
        start, end = (
            code.index(extract_target) + 2,
            code.index(extract_target) + 2 + len(extract_target) - 2,
        )
        refactored = self.do_extract_method(code, start, end, "new_func", similar=True)
        expected = dedent("""\
            class A(object):
                def func(self, val):
                    self[val] = 1
                    var_bb = self.new_func(val)

                def new_func(self, val):
                    return self[val]
        """)

        self.assertEqual(expected, refactored)

    def test_extraction_method_with_global_variable(self):
        code = dedent("""\
            g = None

            def f():
                global g

                g = 2

            f()
            print(g)
        """)
        extract_target = "g = 2"
        start, end = code.index(extract_target), code.index(extract_target) + len(
            extract_target
        )
        refactored = self.do_extract_method(code, start, end, "_g")
        expected = dedent("""\
            g = None

            def f():
                global g

                _g()

            def _g():
                global g
                g = 2

            f()
            print(g)
        """)
        self.assertEqual(expected, refactored)

    def test_extraction_method_with_global_variable_and_global_declaration(self):
        code = dedent("""\
            g = None

            def f():
                global g

                g = 2

            f()
            print(g)
        """)
        start, end = 23, 42
        refactored = self.do_extract_method(code, start, end, "_g")
        expected = dedent("""\
            g = None

            def f():
                _g()

            def _g():
                global g

                g = 2

            f()
            print(g)
        """)
        self.assertEqual(expected, refactored)

    def test_extraction_one_line_with_global_variable_read_only(self):
        code = dedent("""\
            g = None

            def f():
                global g

                a = g

            f()
            print(g)
        """)
        extract_target = "= g"
        start, end = code.index(extract_target) + 2, code.index(extract_target) + 3
        refactored = self.do_extract_method(code, start, end, "_g")
        expected = dedent("""\
            g = None

            def f():
                global g

                a = _g()

            def _g():
                return g

            f()
            print(g)
        """)
        self.assertEqual(expected, refactored)

    @testutils.only_for_versions_higher("3.8")
    def test_extraction_one_line_with_global_variable(self):
        code = dedent("""\
            g = None

            def f():
                global g

                while g := 4:
                    pass

            f()
            print(g)
        """)
        extract_target = "g := 4"
        start, end = code.index(extract_target), code.index(extract_target) + len(
            extract_target
        )
        refactored = self.do_extract_method(code, start, end, "_g")
        expected = dedent("""\
            g = None

            def f():
                global g

                while _g():
                    pass

            def _g():
                global g
                return (g := 4)

            f()
            print(g)
        """)
        self.assertEqual(expected, refactored)

    @testutils.only_for_versions_higher("3.8")
    def test_extraction_one_line_with_global_variable_has_postread(self):
        code = dedent("""\
            g = None

            def f():
                global g

                while g := 4:
                    print(g)

            f()
            print(g)
        """)
        extract_target = "g := 4"
        start, end = code.index(extract_target), code.index(extract_target) + len(
            extract_target
        )
        refactored = self.do_extract_method(code, start, end, "_g")
        expected = dedent("""\
            g = None

            def f():
                global g

                while g := _g():
                    print(g)

            def _g():
                global g
                return (g := 4)

            f()
            print(g)
        """)
        self.assertEqual(expected, refactored)

    def test_extraction_with_nonlocal_variable(self):
        code = dedent("""\
            def outer():
                a = 0
                def inner():
                    nonlocal a
                    a = 3
        """)
        extract_target = "a = 3"
        start, end = code.index(extract_target), code.index(extract_target) + len(
            extract_target
        )
        refactored = self.do_extract_method(code, start, end, "extracted")
        expected = dedent("""\
            def outer():
                a = 0
                def inner():
                    nonlocal a
                    extracted()

                def extracted():
                    nonlocal a
                    a = 3
        """)
        self.assertEqual(expected, refactored)

    def test_extraction_method_with_nonlocal_variable_and_nonlocal_declaration(self):
        code = dedent("""\
            def outer():
                g = None

                def inner():
                    nonlocal g

                    g = 2

                inner()
                print(g)
        """)
        start, end = 52, 78
        refactored = self.do_extract_method(code, start, end, "_g")
        expected = dedent("""\
            def outer():
                g = None

                def inner():
                    _g()

                def _g():
                    nonlocal g

                    g = 2

                inner()
                print(g)
        """)
        self.assertEqual(expected, refactored)

    def test_extraction_one_line_with_nonlocal_variable_read_only(self):
        code = dedent("""\
            def outer():
                g = None

                def inner():
                    nonlocal g

                    a = g

                inner()
                print(g)
        """)
        extract_target = "= g"
        start, end = code.index(extract_target) + 2, code.index(extract_target) + 3
        refactored = self.do_extract_method(code, start, end, "_g")
        expected = dedent("""\
            def outer():
                g = None

                def inner():
                    nonlocal g

                    a = _g()

                def _g():
                    return g

                inner()
                print(g)
        """)
        self.assertEqual(expected, refactored)

    @testutils.only_for_versions_higher("3.8")
    def test_extraction_one_line_with_nonlocal_variable(self):
        code = dedent("""\
            def outer():
                g = None

                def inner():
                    nonlocal g

                    while g := 4:
                        pass

                inner()
                print(g)
        """)
        extract_target = "g := 4"
        start, end = code.index(extract_target), code.index(extract_target) + len(
            extract_target
        )
        refactored = self.do_extract_method(code, start, end, "_g")
        expected = dedent("""\
            def outer():
                g = None

                def inner():
                    nonlocal g

                    while _g():
                        pass

                def _g():
                    nonlocal g
                    return (g := 4)

                inner()
                print(g)
        """)
        self.assertEqual(expected, refactored)

    @testutils.only_for_versions_higher("3.8")
    def test_extraction_one_line_with_nonlocal_variable_has_postread(self):
        code = dedent("""\
            def outer():
                g = None

                def inner():
                    nonlocal g

                    while g := 4:
                        print(g)

                inner()
                print(g)
        """)
        extract_target = "g := 4"
        start, end = code.index(extract_target), code.index(extract_target) + len(
            extract_target
        )
        refactored = self.do_extract_method(code, start, end, "_g")
        expected = dedent("""\
            def outer():
                g = None

                def inner():
                    nonlocal g

                    while g := _g():
                        print(g)

                def _g():
                    nonlocal g
                    return (g := 4)

                inner()
                print(g)
        """)
        self.assertEqual(expected, refactored)

    def test_extract_method_with_nested_double_with_as(self):
        code = dedent("""\
            with open("test") as file1:
                with open("test") as file2:
                    print(file1, file2)
        """)
        start, end = self._convert_line_range_to_offset(code, 3, 4)
        refactored = self.do_extract_method(code, start, end, "extracted", global_=True)
        expected = dedent("""\

            def extracted(file1, file2):
                print(file1, file2)

            with open("test") as file1:
                with open("test") as file2:
                    extracted(file1, file2)
        """)
        self.assertEqual(expected, refactored)

    def test_extract_method_with_double_with_as(self):
        code = dedent("""\
            with open("test") as file1, open("test") as file2:
                print(file1, file2)
        """)
        start, end = self._convert_line_range_to_offset(code, 2, 3)
        refactored = self.do_extract_method(code, start, end, "extracted", global_=True)
        expected = dedent("""\

            def extracted(file1, file2):
                print(file1, file2)

            with open("test") as file1, open("test") as file2:
                extracted(file1, file2)
        """)
        self.assertEqual(expected, refactored)

    def test_extract_method_with_nested_double_with_as_and_misleading_comment(self):
        code = dedent("""\
            with open("test") as file1, open("test") as file2:
                # with in comment
                bar()
        """)
        start, end = self._convert_line_range_to_offset(code, 3, 3)
        refactored = self.do_extract_method(code, start, end, "extracted", global_=True)
        expected = dedent("""\

            def extracted():
                bar()

            with open("test") as file1, open("test") as file2:
                # with in comment
                extracted()
        """)
        self.assertEqual(expected, refactored)

    @testutils.only_for_versions_higher("3.8")
    def test_extract_method_async_with_simple(self):
        code = dedent("""\
            async def afunc():
                async with open("test") as file1:
                    print(file1)
        """)
        start, end = self._convert_line_range_to_offset(code, 2, 3)
        refactored = self.do_extract_method(code, start, end, "extracted", global_=True)
        expected = dedent("""\
            async def afunc():
                extracted()

            def extracted():
                async with open("test") as file1:
                    print(file1)
        """)
        self.assertEqual(expected, refactored)

    @testutils.only_for_versions_higher("3.8")
    def test_extract_method_containing_async_with(self):
        code = dedent("""\
            async def afunc():
                async with open("test") as file1, open("test") as file2:
                    print(file1, file2)
        """)
        start, end = self._convert_line_range_to_offset(code, 3, 3)
        refactored = self.do_extract_method(code, start, end, "extracted", global_=True)
        expected = dedent("""\
            async def afunc():
                async with open("test") as file1, open("test") as file2:
                    extracted(file1, file2)

            def extracted(file1, file2):
                print(file1, file2)
        """)
        self.assertEqual(expected, refactored)

    @testutils.only_for_versions_higher("3.10")
    def test_extract_method_containing_structural_pattern_match(self):
        code = dedent("""\
            match var:
                case Foo("xx"):
                    print(x)
                case Foo(x):
                    print(x)
        """)
        start, end = self._convert_line_range_to_offset(code, 5, 5)
        refactored = self.do_extract_method(code, start, end, "extracted")
        expected = dedent("""\

            def extracted():
                print(x)

            match var:
                case Foo("xx"):
                    print(x)
                case Foo(x):
                    extracted()
        """)
        self.assertEqual(expected, refactored)

    @testutils.only_for_versions_higher("3.10")
    def test_extract_method_containing_structural_pattern_match_2(self):
        code = dedent("""\
            def foo():
                match var:
                    case Foo(x):
                        print(x)
        """)
        start, end = self._convert_line_range_to_offset(code, 4, 4)
        refactored = self.do_extract_method(code, start, end, "extracted")
        expected = dedent("""\
            def foo():
                match var:
                    case Foo(x):
                        extracted(x)

            def extracted(x):
                print(x)
        """)
        self.assertEqual(expected, refactored)

    @testutils.only_for_versions_higher("3.10")
    def test_extract_method_containing_structural_pattern_match_3(self):
        code = dedent("""\
            def foo():
                match var:
                    case {"hello": x} as y:
                        print(x)
        """)
        start, end = self._convert_line_range_to_offset(code, 4, 4)
        refactored = self.do_extract_method(code, start, end, "extracted")
        expected = dedent("""\
            def foo():
                match var:
                    case {"hello": x} as y:
                        extracted(x)

            def extracted(x):
                print(x)
        """)
        self.assertEqual(expected, refactored)



import hashlib
import importlib.util
import json
from typing import Dict

import rope.base.project


def get_version_hash_data(project: rope.base.project.Project) -> Dict[str, str]:
    version_hash_data = dict(
        version_data=f"{rope.VERSION}",
        prefs_data=_get_prefs_data(project),
        schema_file_content=_get_file_content("rope.contrib.autoimport.models"),
    )
    return version_hash_data


def calculate_version_hash(project: rope.base.project.Project) -> str:
    def _merge(hasher, name: str, serialized_data: str):
        hashed_data = hashlib.sha256(serialized_data.encode("utf-8")).hexdigest()
        hasher.update(hashed_data.encode("ascii"))

    hasher = hashlib.sha256()
    for name, data in get_version_hash_data(project).items():
        _merge(hasher, name, data)
    return hasher.hexdigest()


def _get_prefs_data(project) -> str:
    prefs_data = dict(vars(project.prefs))
    del prefs_data["project_opened"]
    del prefs_data["callbacks"]
    del prefs_data["dependencies"]
    return json.dumps(prefs_data, sort_keys=True, indent=2)


def _get_file_content(module_name: str) -> str:
    models_module = importlib.util.find_spec(module_name)
    assert models_module and models_module.loader
    assert isinstance(models_module.loader, importlib.machinery.SourceFileLoader)
    src = models_module.loader.get_source(module_name)
    assert src
    return src



"""Find occurrences of a name in a project.

This module consists of a `Finder` that finds all occurrences of a name
in a project. The `Finder.find_occurrences()` method is a generator that
yields `Occurrence` instances for each occurrence of the name. To create
a `Finder` object, use the `create_finder()` function:

    finder = occurrences.create_finder(project, 'foo', pyname)
    for occurrence in finder.find_occurrences():
        pass

It's possible to filter the occurrences. They can be specified when
calling the `create_finder()` function.

  * `only_calls`: If True, return only those instances where the name is
    a function that's being called.

  * `imports`: If False, don't return instances that are in import
    statements.

  * `unsure`: If a predicate function, return instances where we don't
    know what the name references. It also filters based on the
    predicate function.

  * `docs`: If True, it will search for occurrences in regions normally
    ignored. E.g., strings and comments.

  * `in_hierarchy`: If True, it will find occurrences if the name is in
    the class's hierarchy.

  * `instance`: Used only when you want implicit interfaces to be
    considered.

  * `keywords`: If False, don't return instances that are the names of keyword
    arguments
"""


import contextlib
import re

from rope.base import (
    ast,
    codeanalyze,
    evaluate,
    exceptions,
    pynames,
    pyobjects,
    utils,
    worder,
)


class Finder:
    """For finding occurrences of a name

    The constructor takes a `filters` argument.  It should be a list
    of functions that take a single argument.  For each possible
    occurrence, these functions are called in order with the an
    instance of `Occurrence`:

      * If it returns `None` other filters are tried.
      * If it returns `True`, the occurrence will be a match.
      * If it returns `False`, the occurrence will be skipped.
      * If all of the filters return `None`, it is skipped also.

    """

    def __init__(self, project, name, filters=None, docs=False):
        if filters is None:
            filters = [lambda o: True]
        self.project = project
        self.name = name
        self.docs = docs
        self.filters = filters
        self._textual_finder = _TextualFinder(name, docs=docs)

    def find_occurrences(self, resource=None, pymodule=None):
        """Generate `Occurrence` instances"""
        tools = _OccurrenceToolsCreator(
            self.project, resource=resource, pymodule=pymodule, docs=self.docs
        )
        for offset in self._textual_finder.find_offsets(tools.source_code):
            occurrence = Occurrence(tools, offset)
            for filter in self.filters:
                result = filter(occurrence)
                if result is None:
                    continue
                if result:
                    yield occurrence
                break


def create_finder(
    project,
    name,
    pyname,
    only_calls=False,
    imports=True,
    unsure=None,
    docs=False,
    instance=None,
    in_hierarchy=False,
    keywords=True,
):
    """A factory for `Finder`

    Based on the arguments it creates a list of filters.  `instance`
    argument is needed only when you want implicit interfaces to be
    considered.

    """
    pynames_ = {pyname}
    filters = []
    if only_calls:
        filters.append(CallsFilter())
    if not imports:
        filters.append(NoImportsFilter())
    if not keywords:
        filters.append(NoKeywordsFilter())
    if isinstance(instance, pynames.ParameterName):
        for pyobject in instance.get_objects():
            try:
                pynames_.add(pyobject[name])
            except exceptions.AttributeNotFoundError:
                pass
    for pyname in pynames_:
        filters.append(PyNameFilter(pyname))
        if in_hierarchy:
            filters.append(InHierarchyFilter(pyname))
    if unsure:
        filters.append(UnsureFilter(unsure))
    return Finder(project, name, filters=filters, docs=docs)


class Occurrence:
    def __init__(self, tools, offset):
        self.tools = tools
        self.offset = offset
        self.resource = tools.resource

    @utils.saveit
    def get_word_range(self):
        return self.tools.word_finder.get_word_range(self.offset)

    @utils.saveit
    def get_primary_range(self):
        return self.tools.word_finder.get_primary_range(self.offset)

    @utils.saveit
    def get_pyname(self):
        with contextlib.suppress(exceptions.BadIdentifierError):
            return self.tools.name_finder.get_pyname_at(self.offset)

    @utils.saveit
    def get_primary_and_pyname(self):
        with contextlib.suppress(exceptions.BadIdentifierError):
            return self.tools.name_finder.get_primary_and_pyname_at(self.offset)

    @utils.saveit
    def is_in_import_statement(self):
        return self.tools.word_finder.is_from_statement(
            self.offset
        ) or self.tools.word_finder.is_import_statement(self.offset)

    def is_called(self):
        return self.tools.word_finder.is_a_function_being_called(self.offset)

    def is_defined(self):
        return self.tools.word_finder.is_a_class_or_function_name_in_header(self.offset)

    def is_a_fixed_primary(self):
        return self.tools.word_finder.is_a_class_or_function_name_in_header(
            self.offset
        ) or self.tools.word_finder.is_a_name_after_from_import(self.offset)

    def is_written(self):
        return self.tools.word_finder.is_assigned_here(self.offset)

    def is_unsure(self):
        return unsure_pyname(self.get_pyname())

    def is_function_keyword_parameter(self):
        return self.tools.word_finder.is_function_keyword_parameter(self.offset)

    @property
    @utils.saveit
    def lineno(self):
        offset = self.get_word_range()[0]
        return self.tools.pymodule.lines.get_line_number(offset)


def same_pyname(expected, pyname):
    """Check whether `expected` and `pyname` are the same"""
    if expected is None or pyname is None:
        return False
    if expected == pyname:
        return True
    if not isinstance(
        expected,
        (pynames.ImportedModule, pynames.ImportedName),
    ) and not isinstance(
        pyname,
        (pynames.ImportedModule, pynames.ImportedName),
    ):
        return False
    return (
        expected.get_definition_location() == pyname.get_definition_location()
        and expected.get_object() == pyname.get_object()
    )


def unsure_pyname(pyname, unbound=True):
    """Return `True` if we don't know what this name references"""
    if pyname is None:
        return True
    if unbound and not isinstance(pyname, pynames.UnboundName):
        return False
    if pyname.get_object() == pyobjects.get_unknown():
        return True


class PyNameFilter:
    """For finding occurrences of a name."""

    def __init__(self, pyname):
        self.pyname = pyname

    def __call__(self, occurrence):
        if same_pyname(self.pyname, occurrence.get_pyname()):
            return True


class InHierarchyFilter:
    """Finds the occurrence if the name is in the class's hierarchy."""

    def __init__(self, pyname, implementations_only=False):
        self.pyname = pyname
        self.impl_only = implementations_only
        self.pyclass = self._get_containing_class(pyname)
        if self.pyclass is not None:
            self.name = pyname.get_object().get_name()
            self.roots = self._get_root_classes(self.pyclass, self.name)
        else:
            self.roots = None

    def __call__(self, occurrence):
        if self.roots is None:
            return
        pyclass = self._get_containing_class(occurrence.get_pyname())
        if pyclass is not None:
            roots = self._get_root_classes(pyclass, self.name)
            if self.roots.intersection(roots):
                return True

    def _get_containing_class(self, pyname):
        if isinstance(pyname, pynames.DefinedName):
            scope = pyname.get_object().get_scope()
            parent = scope.parent
            if parent is not None and parent.get_kind() == "Class":
                return parent.pyobject

    def _get_root_classes(self, pyclass, name):
        if self.impl_only and pyclass == self.pyclass:
            return {pyclass}
        result = set()
        for superclass in pyclass.get_superclasses():
            if name in superclass:
                result.update(self._get_root_classes(superclass, name))
        if not result:
            return {pyclass}
        return result


class UnsureFilter:
    """Occurrences where we don't knoow what the name references."""

    def __init__(self, unsure):
        self.unsure = unsure

    def __call__(self, occurrence):
        if occurrence.is_unsure() and self.unsure(occurrence):
            return True


class NoImportsFilter:
    """Don't include import statements as occurrences."""

    def __call__(self, occurrence):
        if occurrence.is_in_import_statement():
            return False


class CallsFilter:
    """Filter out non-call occurrences."""

    def __call__(self, occurrence):
        if not occurrence.is_called():
            return False


class NoKeywordsFilter:
    """Filter out keyword parameters."""

    def __call__(self, occurrence):
        if occurrence.is_function_keyword_parameter():
            return False


class _TextualFinder:
    def __init__(self, name, docs=False):
        self.name = name
        self.docs = docs
        self.comment_pattern = _TextualFinder.any("comment", [r"#[^\n]*"])
        self.string_pattern = _TextualFinder.any(
            "string", [codeanalyze.get_string_pattern()]
        )
        self.f_string_pattern = _TextualFinder.any(
            "fstring", [codeanalyze.get_formatted_string_pattern()]
        )
        self.pattern = self._get_occurrence_pattern(self.name)

    def find_offsets(self, source):
        if not self._fast_file_query(source):
            return
        if self.docs:
            searcher = self._normal_search
        else:
            searcher = self._re_search
        yield from searcher(source)

    def _re_search(self, source):
        for match in self.pattern.finditer(source):
            if match.groupdict()["occurrence"]:
                yield match.start("occurrence")
            elif match.groupdict()["fstring"]:
                f_string = match.groupdict()["fstring"]
                for occurrence_node in self._search_in_f_string(f_string):
                    yield match.start("fstring") + occurrence_node.col_offset

    def _search_in_f_string(self, f_string):
        tree = ast.parse(f_string)
        for node in ast.walk(tree):
            if isinstance(node, ast.Name) and node.id == self.name:
                yield node

    def _normal_search(self, source):
        current = 0
        while True:
            try:
                found = source.index(self.name, current)
                current = found + len(self.name)
                if (found == 0 or not self._is_id_char(source[found - 1])) and (
                    current == len(source) or not self._is_id_char(source[current])
                ):
                    yield found
            except ValueError:
                break

    def _is_id_char(self, c):
        return c.isalnum() or c == "_"

    def _fast_file_query(self, source):
        return self.name in source

    def _get_source(self, resource, pymodule):
        if resource is not None:
            return resource.read()
        else:
            return pymodule.source_code

    def _get_occurrence_pattern(self, name):
        occurrence_pattern = _TextualFinder.any("occurrence", ["\\b" + name + "\\b"])
        pattern = re.compile(
            occurrence_pattern
            + "|"
            + self.comment_pattern
            + "|"
            + self.string_pattern
            + "|"
            + self.f_string_pattern
        )
        return pattern

    @staticmethod
    def any(name, list_):
        return "(?P<%s>" % name + "|".join(list_) + ")"


class _OccurrenceToolsCreator:
    def __init__(self, project, resource=None, pymodule=None, docs=False):
        self.project = project
        self.__resource = resource
        self.__pymodule = pymodule
        self.docs = docs

    @property
    @utils.saveit
    def name_finder(self):
        return evaluate.ScopeNameFinder(self.pymodule)

    @property
    @utils.saveit
    def source_code(self):
        return self.pymodule.source_code

    @property
    @utils.saveit
    def word_finder(self):
        return worder.Worder(self.source_code, self.docs)

    @property
    @utils.saveit
    def resource(self):
        if self.__resource is not None:
            return self.__resource
        if self.__pymodule is not None:
            return self.__pymodule.resource

    @property
    @utils.saveit
    def pymodule(self):
        if self.__pymodule is not None:
            return self.__pymodule
        return self.project.get_pymodule(self.resource)



class RopeError(Exception):
    """Base exception for rope"""


class ResourceNotFoundError(RopeError):
    """Resource not found exception"""


class RefactoringError(RopeError):
    """Errors for performing a refactoring"""


class InterruptedTaskError(RopeError):
    """The task has been interrupted"""


class HistoryError(RopeError):
    """Errors for history undo/redo operations"""


class ModuleNotFoundError(RopeError):
    """Module not found exception"""


class AttributeNotFoundError(RopeError):
    """Attribute not found exception"""


class NameNotFoundError(RopeError):
    """Name not found exception"""


class BadIdentifierError(RopeError):
    """The name cannot be resolved"""


class ModuleSyntaxError(RopeError):
    """Module has syntax errors

    The `filename` and `lineno` fields indicate where the error has
    occurred.

    """

    def __init__(self, filename, lineno, message):
        self.filename = filename
        self.lineno = lineno
        self.message_ = message
        super().__init__(
            f"Syntax error in file <{filename}> line <{lineno}>: {message}"
        )


class ModuleDecodeError(RopeError):
    """Cannot decode module"""

    def __init__(self, filename, message):
        self.filename = filename
        self.message_ = message
        super().__init__(f"Cannot decode file <{filename}>: {message}")



import bisect
import keyword

import rope.base.simplify

MINIMAL_LEN_FOR_AS = 5


def get_name_at(resource, offset):
    source_code = resource.read()
    word_finder = Worder(source_code)
    return word_finder.get_word_at(offset)


class Worder:
    """A class for finding boundaries of words and expressions

    Note that in these methods, offset should be the index of the
    character not the index of the character after it.

    Some of the methods here doesn't exactly do what their name might lead you
    to think they do, these probably should be fixed. Refer to
    ropetest/codeanalyzetest.py for what these methods returns. Note that
    codeanalyzetest.py documents the current behavior, rather than what they
    should've been.
    """

    def __init__(self, code, handle_ignores=False):
        simplified = rope.base.simplify.real_code(code)
        self.code_finder = _RealFinder(simplified, code)
        self.handle_ignores = handle_ignores
        self.code = code

    def _init_ignores(self):
        ignores = rope.base.simplify.ignored_regions(self.code)
        self.dumb_finder = _RealFinder(self.code, self.code)
        self.starts = [ignored[0] for ignored in ignores]
        self.ends = [ignored[1] for ignored in ignores]

    def _context_call(self, name, offset):
        if self.handle_ignores:
            if not hasattr(self, "starts"):
                self._init_ignores()
            start = bisect.bisect(self.starts, offset)
            if start > 0 and offset < self.ends[start - 1]:
                return getattr(self.dumb_finder, name)(offset)
        return getattr(self.code_finder, name)(offset)

    def get_primary_at(self, offset):
        return self._context_call("get_primary_at", offset)

    def get_word_at(self, offset):
        return self._context_call("get_word_at", offset)

    def get_primary_range(self, offset):
        return self._context_call("get_primary_range", offset)

    def get_splitted_primary_before(self, offset):
        return self._context_call("get_splitted_primary_before", offset)

    def get_word_range(self, offset):
        return self._context_call("get_word_range", offset)

    def is_function_keyword_parameter(self, offset):
        return self.code_finder.is_function_keyword_parameter(offset)

    def is_a_class_or_function_name_in_header(self, offset):
        return self.code_finder.is_a_class_or_function_name_in_header(offset)

    def is_from_statement_module(self, offset):
        return self.code_finder.is_from_statement_module(offset)

    def is_from_aliased(self, offset):
        return self.code_finder.is_from_aliased(offset)

    def is_import_statement_aliased_module(self, offset):
        return self.code_finder.is_import_statement_aliased_module(offset)

    def find_parens_start_from_inside(self, offset):
        return self.code_finder.find_parens_start_from_inside(offset)

    def is_a_name_after_from_import(self, offset):
        return self.code_finder.is_a_name_after_from_import(offset)

    def is_from_statement(self, offset):
        return self.code_finder.is_from_statement(offset)

    def get_from_aliased(self, offset):
        return self.code_finder.get_from_aliased(offset)

    def is_import_statement(self, offset):
        return self.code_finder.is_import_statement(offset)

    def is_assigned_here(self, offset):
        return self.code_finder.is_assigned_here(offset)

    def is_a_function_being_called(self, offset):
        return self.code_finder.is_a_function_being_called(offset)

    def get_word_parens_range(self, offset):
        return self.code_finder.get_word_parens_range(offset)

    def is_name_assigned_in_class_body(self, offset):
        return self.code_finder.is_name_assigned_in_class_body(offset)

    def is_on_function_call_keyword(self, offset):
        return self.code_finder.is_on_function_call_keyword(offset)

    def _find_parens_start(self, offset):
        return self.code_finder._find_parens_start(offset)

    def get_parameters(self, first, last):
        return self.code_finder.get_parameters(first, last)

    def get_from_module(self, offset):
        return self.code_finder.get_from_module(offset)

    def is_assigned_in_a_tuple_assignment(self, offset):
        return self.code_finder.is_assigned_in_a_tuple_assignment(offset)

    def get_assignment_type(self, offset):
        return self.code_finder.get_assignment_type(offset)

    def get_function_and_args_in_header(self, offset):
        return self.code_finder.get_function_and_args_in_header(offset)

    def get_lambda_and_args(self, offset):
        return self.code_finder.get_lambda_and_args(offset)

    def find_function_offset(self, offset):
        return self.code_finder.find_function_offset(offset)


class _RealFinder:
    def __init__(self, code, raw):
        self.code = code
        self.raw = raw

    def _find_word_start(self, offset):
        current_offset = offset
        while current_offset >= 0 and self._is_id_char(current_offset):
            current_offset -= 1
        return current_offset + 1

    def _find_word_end(self, offset):
        while offset + 1 < len(self.code) and self._is_id_char(offset + 1):
            offset += 1
        return offset

    def _find_last_non_space_char(self, offset):
        while offset >= 0 and self.code[offset].isspace():
            if self.code[offset] == "\n":
                return offset
            offset -= 1
        return max(-1, offset)

    def get_word_at(self, offset):
        offset = self._get_fixed_offset(offset)
        return self.raw[self._find_word_start(offset) : self._find_word_end(offset) + 1]

    def _get_fixed_offset(self, offset):
        if offset >= len(self.code):
            return offset - 1
        if not self._is_id_char(offset):
            if offset > 0 and self._is_id_char(offset - 1):
                return offset - 1
            if offset < len(self.code) - 1 and self._is_id_char(offset + 1):
                return offset + 1
        return offset

    def _is_id_char(self, offset):
        return self.code[offset].isalnum() or self.code[offset] == "_"

    def _find_string_start(self, offset):
        kind = self.code[offset]
        try:
            return self.code.rindex(kind, 0, offset)
        except ValueError:
            return 0

    def _find_parens_start(self, offset):
        offset = self._find_last_non_space_char(offset - 1)
        while offset >= 0 and self.code[offset] not in "[({":
            if self.code[offset] not in ":,":
                offset = self._find_primary_start(offset)
            offset = self._find_last_non_space_char(offset - 1)
        return offset

    def _find_atom_start(self, offset):
        old_offset = offset
        if self.code[offset] == "\n":
            return offset + 1
        if self.code[offset].isspace():
            offset = self._find_last_non_space_char(offset)
        if self.code[offset] in "'\"":
            return self._find_string_start(offset)
        if self.code[offset] in ")]}":
            return self._find_parens_start(offset)
        if self._is_id_char(offset):
            return self._find_word_start(offset)
        return old_offset

    def _find_primary_without_dot_start(self, offset):
        """It tries to find the undotted primary start

        It is different from `self._get_atom_start()` in that it
        follows function calls, too; such as in ``f(x)``.

        """
        last_atom = offset
        offset = self._find_last_non_space_char(last_atom)
        while offset > 0 and self.code[offset] in ")]":
            last_atom = self._find_parens_start(offset)
            offset = self._find_last_non_space_char(last_atom - 1)
        if offset >= 0 and (self.code[offset] in "\"'})]" or self._is_id_char(offset)):
            atom_start = self._find_atom_start(offset)
            if not keyword.iskeyword(self.code[atom_start : offset + 1]) or (
                offset + 1 < len(self.code) and self._is_id_char(offset + 1)
            ):
                return atom_start
        return last_atom

    def _find_primary_start(self, offset):
        if offset >= len(self.code):
            offset = len(self.code) - 1
        if self.code[offset] != ".":
            offset = self._find_primary_without_dot_start(offset)
        else:
            offset = offset + 1
        while offset > 0:
            prev = self._find_last_non_space_char(offset - 1)
            if offset <= 0 or self.code[prev] != ".":
                break

            # Check if relative import
            # XXX: Looks like a hack...
            prev_word_end = self._find_last_non_space_char(prev - 1)
            if self.code[prev_word_end - 3 : prev_word_end + 1] == "from":
                offset = prev
                break

            offset = self._find_primary_without_dot_start(prev - 1)
            if not self._is_id_char(offset):
                break

        return offset

    def get_primary_at(self, offset):
        offset = self._get_fixed_offset(offset)
        start, end = self.get_primary_range(offset)
        return self.raw[start:end].strip()

    def get_splitted_primary_before(self, offset):
        """returns expression, starting, starting_offset

        This function is used in `rope.codeassist.assist` function.
        """
        if offset == 0:
            return ("", "", 0)
        end = offset - 1
        word_start = self._find_atom_start(end)
        real_start = self._find_primary_start(end)
        if self.code[word_start:offset].strip() == "":
            word_start = end
        if self.code[end].isspace():
            word_start = end
        if self.code[real_start:word_start].strip() == "":
            real_start = word_start
        if real_start == word_start == end and not self._is_id_char(end):
            return ("", "", offset)
        if real_start == word_start:
            return ("", self.raw[word_start:offset], word_start)
        else:
            if self.code[end] == ".":
                return (self.raw[real_start:end], "", offset)
            last_dot_position = word_start
            if self.code[word_start] != ".":
                last_dot_position = self._find_last_non_space_char(word_start - 1)
            last_char_position = self._find_last_non_space_char(last_dot_position - 1)
            if self.code[word_start].isspace():
                word_start = offset
            return (
                self.raw[real_start : last_char_position + 1],
                self.raw[word_start:offset],
                word_start,
            )

    def _get_line_start(self, offset):
        try:
            return self.code.rindex("\n", 0, offset + 1)
        except ValueError:
            return 0

    def _get_line_end(self, offset):
        try:
            return self.code.index("\n", offset)
        except ValueError:
            return len(self.code)

    def is_name_assigned_in_class_body(self, offset):
        word_start = self._find_word_start(offset - 1)
        word_end = self._find_word_end(offset) + 1
        if "." in self.code[word_start:word_end]:
            return False
        line_start = self._get_line_start(word_start)
        line = self.code[line_start:word_start].strip()
        return not line and self.get_assignment_type(offset) == "="

    def is_a_class_or_function_name_in_header(self, offset):
        word_start = self._find_word_start(offset - 1)
        line_start = self._get_line_start(word_start)
        prev_word = self.code[line_start:word_start].strip()
        return prev_word in ["def", "class"]

    def _find_first_non_space_char(self, offset):
        if offset >= len(self.code):
            return len(self.code)
        while offset < len(self.code) and self.code[offset].isspace():
            if self.code[offset] == "\n":
                return offset
            offset += 1
        return offset

    def is_a_function_being_called(self, offset):
        word_end = self._find_word_end(offset) + 1
        next_char = self._find_first_non_space_char(word_end)
        return (
            next_char < len(self.code)
            and self.code[next_char] == "("
            and not self.is_a_class_or_function_name_in_header(offset)
        )

    def _find_import_end(self, start):
        return self._get_line_end(start)

    def is_import_statement(self, offset):
        try:
            last_import = self.code.rindex("import ", 0, offset)
        except ValueError:
            return False
        line_start = self._get_line_start(last_import)
        return (
            self._find_import_end(last_import + 7) >= offset
            and self._find_word_start(line_start) == last_import
        )

    def is_from_statement(self, offset):
        try:
            last_from = self.code.rindex("from ", 0, offset)
            from_import = self.code.index(" import ", last_from)
            from_names = from_import + 8
        except ValueError:
            return False
        from_names = self._find_first_non_space_char(from_names)
        return self._find_import_end(from_names) >= offset

    def is_from_statement_module(self, offset):
        if offset >= len(self.code) - 1:
            return False
        stmt_start = self._find_primary_start(offset)
        line_start = self._get_line_start(stmt_start)
        prev_word = self.code[line_start:stmt_start].strip()
        return prev_word == "from"

    def is_import_statement_aliased_module(self, offset):
        if not self.is_import_statement(offset):
            return False
        try:
            line_start = self._get_line_start(offset)
            import_idx = self.code.rindex("import", line_start, offset)
            imported_names = import_idx + 7
        except ValueError:
            return False
        # Check if the offset is within the imported names
        if (
            imported_names - 1 > offset
            or self._find_import_end(imported_names) < offset
        ):
            return False
        try:
            end = self._find_import_main_part_end(offset)
            if not self._has_enough_len_for_as(end):
                return False
            as_end = min(self._find_word_end(end + 1), len(self.code))
            as_start = self._find_word_start(as_end)
            return self.code[as_start : as_end + 1] == "as"
        except ValueError:
            return False

    def _has_enough_len_for_as(self, end):
        return len(self.code) > end + MINIMAL_LEN_FOR_AS

    def _find_import_main_part_end(self, offset):
        end = self._find_word_end(offset)
        while len(self.code) > end + 2 and self.code[end + 1] == ".":
            end = self._find_word_end(end + 2)
        return end

    def is_a_name_after_from_import(self, offset):
        try:
            if len(self.code) > offset and self.code[offset] == "\n":
                line_start = self._get_line_start(offset - 1)
            else:
                line_start = self._get_line_start(offset)
            last_from = self.code.rindex("from ", line_start, offset)
            from_import = self.code.index(" import ", last_from)
            from_names = from_import + 8
        except ValueError:
            return False
        if from_names - 1 > offset:
            return False
        return self._find_import_end(from_names) >= offset

    def get_from_module(self, offset):
        try:
            last_from = self.code.rindex("from ", 0, offset)
            import_offset = self.code.index(" import ", last_from)
            end = self._find_last_non_space_char(import_offset)
            return self.get_primary_at(end)
        except ValueError:
            pass

    def is_from_aliased(self, offset):
        if not self.is_a_name_after_from_import(offset):
            return False
        try:
            end = self._find_word_end(offset)
            as_end = min(self._find_word_end(end + 1), len(self.code))
            as_start = self._find_word_start(as_end)
            return self.code[as_start : as_end + 1] == "as"
        except ValueError:
            return False

    def get_from_aliased(self, offset):
        try:
            end = self._find_word_end(offset)
            as_ = self._find_word_end(end + 1)
            alias = self._find_word_end(as_ + 1)
            start = self._find_word_start(alias)
            return self.raw[start : alias + 1]
        except ValueError:
            pass

    def is_function_keyword_parameter(self, offset):
        word_end = self._find_word_end(offset)
        if word_end + 1 == len(self.code):
            return False
        next_char = self._find_first_non_space_char(word_end + 1)
        equals = self.code[next_char : next_char + 2]
        if equals == "==" or not equals.startswith("="):
            return False
        word_start = self._find_word_start(offset)
        prev_char = self._find_last_non_space_char(word_start - 1)
        return prev_char - 1 >= 0 and self.code[prev_char] in ",("

    def is_on_function_call_keyword(self, offset):
        stop = self._get_line_start(offset)
        if self._is_id_char(offset):
            offset = self._find_word_start(offset) - 1
        offset = self._find_last_non_space_char(offset)
        if offset <= stop or self.code[offset] not in "(,":
            return False
        parens_start = self.find_parens_start_from_inside(offset)
        return stop < parens_start

    def find_parens_start_from_inside(self, offset):
        stop = self._get_line_start(offset)
        while offset > stop:
            if self.code[offset] == "(":
                break
            if self.code[offset] != ",":
                offset = self._find_primary_start(offset)
            offset -= 1
        return max(stop, offset)

    def is_assigned_here(self, offset):
        return self.get_assignment_type(offset) is not None

    def get_assignment_type(self, offset):
        # XXX: does not handle tuple assignments
        word_end = self._find_word_end(offset)
        next_char = self._find_first_non_space_char(word_end + 1)
        single = self.code[next_char : next_char + 1]
        double = self.code[next_char : next_char + 2]
        triple = self.code[next_char : next_char + 3]
        if double not in ("==", "<=", ">=", "!="):
            for op in [single, double, triple]:
                if op.endswith("="):
                    return op

    def get_primary_range(self, offset):
        start = self._find_primary_start(offset)
        end = self._find_word_end(offset) + 1
        return (start, end)

    def get_word_range(self, offset):
        offset = max(0, offset)
        start = self._find_word_start(offset)
        end = self._find_word_end(offset) + 1
        return (start, end)

    def get_word_parens_range(self, offset, opening="(", closing=")"):
        end = self._find_word_end(offset)
        start_parens = self.code.index(opening, end)
        index = start_parens
        open_count = 0
        while index < len(self.code):
            if self.code[index] == opening:
                open_count += 1
            if self.code[index] == closing:
                open_count -= 1
            if open_count == 0:
                return (start_parens, index + 1)
            index += 1
        return (start_parens, index)

    def get_parameters(self, first, last):
        keywords = []
        args = []
        current = self._find_last_non_space_char(last - 1)
        while current > first:
            primary_start = current
            current = self._find_primary_start(current)
            while current != first and (
                self.code[current] not in "=," or self.code[current - 1] in "=!<>"
            ):
                current = self._find_last_non_space_char(current - 1)
            primary = self.raw[current + 1 : primary_start + 1].strip()
            if self.code[current] == "=":
                primary_start = current - 1
                current -= 1
                while current != first and self.code[current] not in ",":
                    current = self._find_last_non_space_char(current - 1)
                param_name = self.raw[current + 1 : primary_start + 1].strip()
                keywords.append((self.__strip_type_hint(param_name), primary))
            else:
                args.append(self.__strip_type_hint(primary))
            current = self._find_last_non_space_char(current - 1)
        args.reverse()
        keywords.reverse()
        return args, keywords

    def __strip_type_hint(self, name):
        return name.split(":", 1)[0]

    def is_assigned_in_a_tuple_assignment(self, offset):
        start = self._get_line_start(offset)
        end = self._get_line_end(offset)
        primary_start = self._find_primary_start(offset)
        primary_end = self._find_word_end(offset)

        prev_char_offset = self._find_last_non_space_char(primary_start - 1)
        next_char_offset = self._find_first_non_space_char(primary_end + 1)
        next_char = prev_char = ""
        if prev_char_offset >= start:
            prev_char = self.code[prev_char_offset]
        if next_char_offset < end:
            next_char = self.code[next_char_offset]
        try:
            equals_offset = self.code.index("=", start, end)
        except ValueError:
            return False
        if prev_char not in "(," and next_char not in ",)":
            return False
        parens_start = self.find_parens_start_from_inside(offset)
        # XXX: only handling (x, y) = value
        return offset < equals_offset and self.code[start:parens_start].strip() == ""

    def get_function_and_args_in_header(self, offset):
        offset = self.find_function_offset(offset)
        lparens, rparens = self.get_word_parens_range(offset)
        return self.raw[offset : rparens + 1]

    def find_function_offset(self, offset, definition="def "):
        while True:
            offset = self.code.index(definition, offset)
            if offset == 0 or not self._is_id_char(offset - 1):
                break
            offset += 1
        def_ = offset + 4
        return self._find_first_non_space_char(def_)

    def get_lambda_and_args(self, offset):
        offset = self.find_function_offset(offset, definition="lambda ")
        lparens, rparens = self.get_word_parens_range(offset, opening=" ", closing=":")
        return self.raw[offset : rparens + 1]



import unittest

from rope.base.oi import memorydb, objectdb
from ropetest import testutils


def _do_for_all_dbs(function):
    def called(self):
        for db in self.dbs:
            function(self, db)

    return called


class _MockValidation:
    def is_value_valid(self, value):
        return value != -1

    def is_more_valid(self, new, old):
        return new != -1

    def is_file_valid(self, path):
        return path != "invalid"

    def is_scope_valid(self, path, key):
        return path != "invalid" and key != "invalid"


class _MockFileListObserver:

    log = ""

    def added(self, path):
        self.log += "added %s " % path

    def removed(self, path):
        self.log += "removed %s " % path


class ObjectDBTest(unittest.TestCase):
    def setUp(self):
        super().setUp()
        self.project = testutils.sample_project()
        validation = _MockValidation()
        self.dbs = [objectdb.ObjectDB(memorydb.MemoryDB(self.project), validation)]

    def tearDown(self):
        for db in self.dbs:
            db.write()
        testutils.remove_project(self.project)
        super().tearDown()

    @_do_for_all_dbs
    def test_simple_per_name(self, db):
        db.add_pername("file", "key", "name", 1)
        self.assertEqual(1, db.get_pername("file", "key", "name"))

    @_do_for_all_dbs
    def test_simple_per_name_does_not_exist(self, db):
        self.assertEqual(None, db.get_pername("file", "key", "name"))

    @_do_for_all_dbs
    def test_simple_per_name_after_syncing(self, db):
        db.add_pername("file", "key", "name", 1)
        db.write()
        self.assertEqual(1, db.get_pername("file", "key", "name"))

    @_do_for_all_dbs
    def test_getting_returned(self, db):
        db.add_callinfo("file", "key", (1, 2), 3)
        self.assertEqual(3, db.get_returned("file", "key", (1, 2)))

    @_do_for_all_dbs
    def test_getting_returned_when_does_not_match(self, db):
        db.add_callinfo("file", "key", (1, 2), 3)
        self.assertEqual(None, db.get_returned("file", "key", (1, 1)))

    @_do_for_all_dbs
    def test_getting_call_info(self, db):
        db.add_callinfo("file", "key", (1, 2), 3)

        call_infos = list(db.get_callinfos("file", "key"))
        self.assertEqual(1, len(call_infos))
        self.assertEqual((1, 2), call_infos[0].get_parameters())
        self.assertEqual(3, call_infos[0].get_returned())

    @_do_for_all_dbs
    def test_invalid_per_name(self, db):
        db.add_pername("file", "key", "name", -1)
        self.assertEqual(None, db.get_pername("file", "key", "name"))

    @_do_for_all_dbs
    def test_overwriting_per_name(self, db):
        db.add_pername("file", "key", "name", 1)
        db.add_pername("file", "key", "name", 2)
        self.assertEqual(2, db.get_pername("file", "key", "name"))

    @_do_for_all_dbs
    def test_not_overwriting_with_invalid_per_name(self, db):
        db.add_pername("file", "key", "name", 1)
        db.add_pername("file", "key", "name", -1)
        self.assertEqual(1, db.get_pername("file", "key", "name"))

    @_do_for_all_dbs
    def test_getting_invalid_returned(self, db):
        db.add_callinfo("file", "key", (1, 2), -1)
        self.assertEqual(None, db.get_returned("file", "key", (1, 2)))

    @_do_for_all_dbs
    def test_not_overwriting_with_invalid_returned(self, db):
        db.add_callinfo("file", "key", (1, 2), 3)
        db.add_callinfo("file", "key", (1, 2), -1)
        self.assertEqual(3, db.get_returned("file", "key", (1, 2)))

    @_do_for_all_dbs
    def test_get_files(self, db):
        db.add_callinfo("file1", "key", (1, 2), 3)
        db.add_callinfo("file2", "key", (1, 2), 3)
        self.assertEqual({"file1", "file2"}, set(db.get_files()))

    @_do_for_all_dbs
    def test_validating_files(self, db):
        db.add_callinfo("invalid", "key", (1, 2), 3)
        db.validate_files()
        self.assertEqual(0, len(db.get_files()))

    @_do_for_all_dbs
    def test_validating_file_for_scopes(self, db):
        db.add_callinfo("file", "invalid", (1, 2), 3)
        db.validate_file("file")
        self.assertEqual(1, len(db.get_files()))
        self.assertEqual(0, len(list(db.get_callinfos("file", "invalid"))))

    @_do_for_all_dbs
    def test_validating_file_moved(self, db):
        db.add_callinfo("file", "key", (1, 2), 3)

        db.file_moved("file", "newfile")
        self.assertEqual(1, len(db.get_files()))
        self.assertEqual(1, len(list(db.get_callinfos("newfile", "key"))))

    @_do_for_all_dbs
    def test_using_file_list_observer(self, db):
        db.add_callinfo("invalid", "key", (1, 2), 3)
        observer = _MockFileListObserver()
        db.add_file_list_observer(observer)
        db.validate_files()
        self.assertEqual("removed invalid ", observer.log)

    @_do_for_all_dbs
    def test_legacy_serialization(self, db):
        import pickle

        db.add_callinfo("file", "key", (1, 2), 3)
        db.add_pername("file", "key", "name", 1)
        scope_info = db._get_scope_info("file", "key")

        pickled_data = b'\x80\x04\x95D\x00\x00\x00\x00\x00\x00\x00\x8c\x15rope.base.oi.memorydb\x94\x8c\tScopeInfo\x94\x93\x94)\x81\x94}\x94K\x01K\x02\x86\x94K\x03s}\x94\x8c\x04name\x94K\x01s\x86\x94b.'  # noqa

        assert pickle.loads(pickled_data).call_info == scope_info.call_info
        assert pickle.loads(pickled_data).per_name == scope_info.per_name

    @_do_for_all_dbs
    def test_new_pickle_serialization(self, db):
        import pickle

        db.add_callinfo("file", "key", (1, 2), 3)
        db.add_pername("file", "key", "name", 1)
        scope_info = db._get_scope_info("file", "key")

        serialized = pickle.dumps(scope_info)

        rehydrated_data = pickle.loads(serialized)
        assert rehydrated_data.call_info == scope_info.call_info
        assert rehydrated_data.per_name == scope_info.per_name

    @_do_for_all_dbs
    def test_new_json_serialization(self, db):
        import json

        from rope.base.oi.memorydb import ScopeInfo

        db.add_callinfo("file", "key", (1, 2), 3)
        db.add_pername("file", "key", "name", 1)
        scope_info = db._get_scope_info("file", "key")

        data = {"inside": [scope_info], "other": scope_info, "things": [1, 2, 3]}

        def object_hook(o):
            if o.get("$") == "ScopeInfo":
                new_o = ScopeInfo.__new__(ScopeInfo)
                new_o.__setstate__(o)
                return new_o
            return o

        serialized = json.dumps(data, default=lambda o: o.__getstate__())
        rehydrated_data = json.loads(serialized, object_hook=object_hook)

        rehydrated_scope_info = rehydrated_data["inside"][0]
        assert isinstance(rehydrated_scope_info, ScopeInfo)
        assert rehydrated_scope_info.call_info == scope_info.call_info
        assert rehydrated_scope_info.per_name == scope_info.per_name



class IResolver:
    def __call__(self, hint, pyobject):
        """
        :param hint: For example "List[int]" or "(Foo, Bar) -> Baz" or simple "Foo"
        :type hint: str
        :type pyobject: rope.base.pyobjects.PyDefinedObject | rope.base.pyobjects.PyObject
        :rtype: rope.base.pyobjects.PyDefinedObject | rope.base.pyobjects.PyObject or None
        """
        raise NotImplementedError



import unittest

from rope.base import simplify


class SimplifyTest(unittest.TestCase):
    def test_trivial_case(self):
        self.assertEqual("", simplify.real_code(""))

    def test_empty_strs(self):
        code = 's = ""\n'
        self.assertEqual(code, simplify.real_code(code))

    def test_blanking_strs(self):
        code = 's = "..."\n'
        self.assertEqual('s = "   "\n', simplify.real_code(code))

    def test_changing_to_double_quotes(self):
        code = "s = ''\n"
        self.assertEqual('s = ""\n', simplify.real_code(code))

    def test_changing_to_double_quotes2(self):
        code = 's = """\n"""\n'
        self.assertEqual('s = "     "\n', simplify.real_code(code))

    def test_removing_comments(self):
        code = "# c\n"
        self.assertEqual("   \n", simplify.real_code(code))

    def test_removing_comments_that_contain_strings(self):
        code = '# "c"\n'
        self.assertEqual("     \n", simplify.real_code(code))

    def test_removing_strings_containing_comments(self):
        code = '"#c"\n'
        self.assertEqual('"  "\n', simplify.real_code(code))

    def test_joining_implicit_continuations(self):
        code = "(\n)\n"
        self.assertEqual("( )\n", simplify.real_code(code))

    def test_joining_explicit_continuations(self):
        code = "1 + \\\n 2\n"
        self.assertEqual("1 +    2\n", simplify.real_code(code))

    def test_replacing_tabs(self):
        code = "1\t+\t2\n"
        self.assertEqual("1 + 2\n", simplify.real_code(code))

    def test_replacing_semicolons(self):
        code = "a = 1;b = 2\n"
        self.assertEqual("a = 1\nb = 2\n", simplify.real_code(code))

    def test_simplifying_f_string(self):
        code = 's = f"..{hello}.."\n'
        self.assertEqual('s = f"..{hello}.."\n', simplify.real_code(code))

    def test_simplifying_f_string_containing_quotes(self):
        code = """s = f"..'{hello}'.."\n"""
        self.assertEqual("""s = f"..'{hello}'.."\n""", simplify.real_code(code))

    def test_simplifying_uppercase_f_string_containing_quotes(self):
        code = """s = Fr"..'{hello}'.."\n"""
        self.assertEqual("""s = Fr"..'{hello}'.."\n""", simplify.real_code(code))



from abc import ABC, abstractmethod
from typing import Optional, Sequence

from rope.base import exceptions, utils


class BaseJobSet(ABC):
    name: str = ""
    job_name: str = ""

    @abstractmethod
    def started_job(self, name: str) -> None:
        pass

    @abstractmethod
    def finished_job(self) -> None:
        pass

    @abstractmethod
    def check_status(self) -> None:
        pass

    @utils.deprecated("Just use JobSet.job_name attribute/property instead")
    def get_active_job_name(self) -> str:
        pass

    @abstractmethod
    def get_percent_done(self) -> Optional[float]:
        pass

    @utils.deprecated("Just use JobSet.name attribute/property instead")
    def get_name(self) -> str:
        pass

    @abstractmethod
    def increment(self) -> None:
        """
        Increment the number of tasks to complete.

        This is used if the number is not known ahead of time.
        """
        pass


class BaseTaskHandle(ABC):
    @abstractmethod
    def stop(self) -> None:
        pass

    @abstractmethod
    def current_jobset(self) -> Optional[BaseJobSet]:
        pass

    @abstractmethod
    def add_observer(self) -> None:
        pass

    @abstractmethod
    def is_stopped(self) -> bool:
        pass

    @abstractmethod
    def get_jobsets(self) -> Sequence[BaseJobSet]:
        pass

    def create_jobset(
        self, name: str = "JobSet", count: Optional[int] = None
    ) -> BaseJobSet:
        pass

    def _inform_observers(self) -> None:
        pass


class TaskHandle(BaseTaskHandle):
    def __init__(self, name="Task", interrupts=True):
        """Construct a TaskHandle

        If `interrupts` is `False` the task won't be interrupted by
        calling `TaskHandle.stop()`.

        """
        self.name = name
        self.interrupts = interrupts
        self.stopped = False
        self.job_sets = []
        self.observers = []

    def stop(self):
        """Interrupts the refactoring"""
        if self.interrupts:
            self.stopped = True
            self._inform_observers()

    def current_jobset(self):
        """Return the current `JobSet`"""
        if self.job_sets:
            return self.job_sets[-1]

    def add_observer(self, observer):
        """Register an observer for this task handle

        The observer is notified whenever the task is stopped or
        a job gets finished.

        """
        self.observers.append(observer)

    def is_stopped(self):
        return self.stopped

    def get_jobsets(self):
        return self.job_sets

    def create_jobset(self, name="JobSet", count=None):
        result = JobSet(self, name=name, count=count)
        self.job_sets.append(result)
        self._inform_observers()
        return result

    def _inform_observers(self):
        for observer in list(self.observers):
            observer()


class JobSet(BaseJobSet):
    def __init__(self, handle, name, count):
        self.handle = handle
        self.name = name
        self.count = count
        self.done = 0
        self.job_name = None

    def started_job(self, name):
        self.check_status()
        self.job_name = name
        self.handle._inform_observers()

    def finished_job(self):
        self.check_status()
        self.done += 1
        self.handle._inform_observers()
        self.job_name = None

    def check_status(self):
        if self.handle.is_stopped():
            raise exceptions.InterruptedTaskError()

    @utils.deprecated("Just use JobSet.job_name attribute/property instead")
    def get_active_job_name(self):
        return self.job_name

    def get_percent_done(self):
        if self.count is not None and self.count > 0:
            percent = self.done * 100 // self.count
            return min(percent, 100)

    @utils.deprecated("Just use JobSet.name attribute/property instead")
    def get_name(self):
        return self.name

    def increment(self):
        self.count += 1


class NullTaskHandle(BaseTaskHandle):
    def is_stopped(self):
        return False

    def stop(self):
        pass

    def create_jobset(self, *args, **kwds):
        return NullJobSet()

    def get_jobsets(self):
        return []

    def add_observer(self, observer):
        pass

    def current_jobset(self) -> None:
        """Return the current `JobSet`"""
        return None


class NullJobSet(BaseJobSet):
    def started_job(self, name):
        pass

    def finished_job(self):
        pass

    def check_status(self):
        pass

    @utils.deprecated("Just use JobSet.job_name attribute/property instead")
    def get_active_job_name(self):
        pass

    def get_percent_done(self):
        pass

    @utils.deprecated("Just use JobSet.name attribute/property instead")
    def get_name(self):
        pass

    def increment(self):
        pass


DEFAULT_TASK_HANDLE = NullTaskHandle()
DEFAULT_JOB_SET = NullJobSet()
